import ast
import importlib
import inspect
import logging
import shutil
import traceback
from types import ModuleType
from typing import (
    Any,
    Callable,
    Collection,
    Dict,
    get_args,
    Iterable,
    List,
    Optional,
    Type,
    Union,
)

import docstring_parser
from docutils import nodes, statemachine  # type: ignore
from docutils.parsers.rst import Directive  # type: ignore
from sphinx import application  # type: ignore
from sphinx.util import nodes as sphinx_nodes  # type: ignore

from jobmon.client.api import Tool
from jobmon.client.task import Task
from jobmon.client.workflow import Workflow

SIMPLE_TYPES = {str, int, float, bool}
BUILT_IN_COLLECTIONS = {list, tuple, set}
OPTIONAL_TYPES = {Optional, type(None)}

TASK_RUNNER_NAME = "worker_node_entry_point"
TASK_RUNNER_SUB_COMMAND = "task_generator"
HELP_TEXT_INTRO_FORMAT = "Command Line Documentation for {full_path}"

SERIALIZED_EMPTY_STRING = '""'

logger = logging.getLogger(__name__)


def create_task_name(
    kwargs_for_name: dict,
    prefix: str = "",
    name_func: Optional[Callable] = None,
) -> str:
    """Create a task name from the kwargs."""
    if name_func:
        return name_func(prefix, kwargs_for_name)
    else:
        # use FHS default if no name_func is provided
        # Handle the case where we have an empty string placeholder in the name by making it
        # empty in the name; Jobmon will not accept quotes in the name
        for key, value in kwargs_for_name.items():
            if value == SERIALIZED_EMPTY_STRING:
                kwargs_for_name[key] = ""

        name = prefix
        for item_name, value in kwargs_for_name.items():
            # trim leading and ending single quote added by make_cli_argument_string
            if value[0] == "'" and value[-1] == "'":
                value = value[1:-1]
            # trim leading [ and ending ] when converting a list to string
            if value[0] == "[" and value[-1] == "]":
                value = value[1:-1]
            # remove illegal characters: '/\\'\" ' from name
            value = (
                value.replace("/", "_")
                .replace("\\", "_")
                .replace('"', "_")
                .replace("'", "_")
                .replace(" ", "_")
            )
            name += f":{item_name}={value}"
        return name


def _is_multidimensional_type(type_hint: Any) -> bool:
    # Check if the type hint is a List
    if (
        hasattr(type_hint, "__origin__")
        and type_hint.__origin__ in BUILT_IN_COLLECTIONS  # type: ignore
    ):
        # Get the arguments inside the List
        args = get_args(type_hint)
        # Check if the first argument is also a List
        if (
            args
            and hasattr(args[0], "__origin__")
            and args[0].__origin__ in BUILT_IN_COLLECTIONS  # type: ignore
        ):
            return True
    return False


def _find_executable_path(executable_name: str) -> str:
    path = shutil.which(executable_name)
    if path is not None:
        return path
    else:
        raise FileNotFoundError(
            f"The executable '{executable_name}' was not found in the system PATH."
        )


def _is_unannotated_built_in_collection_type(obj_type: Any) -> bool:
    """Return whether the ``obj_type`` is an unannotated, built-in collection type.

    Ex. ``list`` would return ``True``, ``tuple`` would return ``True``, ``list[str]``
    would return ``False``.
    """
    return obj_type in BUILT_IN_COLLECTIONS


def _get_collection_type(obj_type: Any) -> Optional[Type]:
    """Return the type of collection.

    To check, we extract the ``__origin__`` component of the ``obj_type`` -- this would be
    the type of collection (list, tuple, set).

    Ex. ``list[int]`` would return ``list``, ``tuple[str]`` would return ``tuple``,
    ``float`` would return ``None`` (because it is not a collection).
    """
    # This catches unannotated built-in collection types (``list``, ``tuple``, ``set``)
    if obj_type in BUILT_IN_COLLECTIONS:
        return obj_type

    return getattr(obj_type, "__origin__", None)


def _get_generic_type_parameters(obj_type: Any) -> Any:
    """Return the parameters for a generic type.

    The results are a tuple of types. For example, a ``list[int]`` returns ``(int,)``, a
    ``tuple[int, str]``, returns ``(int, str)``, and a ``Union[str, None]`` returns
    ``(str, NoneType)``.
    """
    return getattr(obj_type, "__args__", None)


def _is_annotated_built_in_collection_type(obj_type: Any) -> bool:
    """Return whether the type is a properly annotated built-in collection type."""
    collection_type = _get_collection_type(obj_type)
    collection_items_type = _get_generic_type_parameters(obj_type)

    return collection_type in BUILT_IN_COLLECTIONS and collection_items_type is not None


def _zip_collection_items_and_types(
    obj: Collection[str], collection_items_type: tuple
) -> Any:
    """Return a list of tuples mapping collection items and their types.

    Ex. ``obj`` is ``[1, 2, 3]`` and ``collection_items_type`` is ``(int,)`` would return
    ``[(1, int), (2, int), (3, int)]``.
    """
    # If the collection has a single internal type annotation, or if the second
    # internal type annotation is the Ellipsis (``...``), match the first (only) internal
    # annotation to all items in the collection
    if len(collection_items_type) == 1 or collection_items_type[1] == Ellipsis:
        return [(item, collection_items_type[0]) for item in obj]

    # If the collection has multiple internal type annotations but the object
    # provided is a collection of a different length, raise an error
    if len(obj) != len(collection_items_type):
        raise TypeError(
            f"Expected a collection of internal type ``{collection_items_type}`` but got "
            f"``{obj}``. Note the mismatched lengths."
        )

    # Otherwise, match each item of the collection to the type of the corresponding internal
    # type annotation
    return list(zip(obj, collection_items_type))


def is_optional_type(obj_type: Any) -> bool:
    """Return whether the ``obj_type`` is an Optional type.

    Ex. ``Optional[str]`` would return ``True``, ``NoneType`` would return ``True``,
    ``str | None`` would return ``True``, ``str`` would return ``False``.
    """
    matches: List[bool] = []

    # This is true if ``obj_type=NoneType``
    exact_type_match = obj_type in OPTIONAL_TYPES
    matches.append(exact_type_match)

    # This is true if ``obj_type=Optional[str]``. The _name attribute is set by
    # ``typing.Optional``
    possible_annotation = getattr(obj_type, "_name", None)
    annotated_type_match = possible_annotation is not None and any(
        possible_annotation in annotation for annotation in map(str, OPTIONAL_TYPES)
    )
    matches.append(annotated_type_match)

    # This is true if ``obj_type=str | None`` (Note that this is represented as a collection
    # so we can access it using ``_get_generic_type_parameters``)
    possible_union_type = _get_generic_type_parameters(obj_type)
    if possible_union_type is not None:
        possible_union_annotation = any(
            annotation in OPTIONAL_TYPES for annotation in possible_union_type
        )
        matches.append(possible_union_annotation)

    return any(matches)


def get_optional_type_parameter(obj_type: Type) -> Type:
    """Return the type inside an optional."""
    # theoretically a tuple with NoneType and the other type
    types = _get_generic_type_parameters(obj_type)
    if types is not None:
        for t in types:
            if t != type(None):  # noqa: E721
                return t

    raise TypeError(
        f"Looks like we couldn't extract a non-None type from optional-type {obj_type}. This "
        "shouldn't have happend."
    )


def _clean_arg_name(arg_name: str) -> str:
    """Remove the ``--`` from the arg name and convert dashes to underscores."""
    return arg_name.replace("--", "").replace("-", "_")


def _get_short_description(task_function_docstring: docstring_parser.Docstring) -> str:
    short_description = ""
    if task_function_docstring.short_description:
        short_description += "\n" + task_function_docstring.short_description
    return short_description


def make_cli_argument_string(arg_value: Union[str, List]) -> str:
    """Make a CLI argument string from an argument name and value.

    For string, return itself.
    For list, say ["a", "b", "c"], return string '[a,b,c]'
    """
    if isinstance(arg_value, list):
        no_space_string = ",".join(arg_value)
        # we can not take single quote in the string, throw an error
        if "'" in no_space_string:
            raise ValueError(
                f"This version of TaskGenertor cannot serialize list with single quote in it: "
                f"{arg_value}."
            )
        return f"'[{no_space_string}]'"

    return f"'{arg_value}'"


class TaskGenerator:
    """Class for auto-generating jobmon tasks from a python function.

    Arguments on the task_function are inspected to generate the task template, and type
    annotations are used for serializing and deserializing arguments to strings passed on the
    command line.

    The TaskGenerator has built in serializers for:

        * str, int, bool, float
        * list, tuple, set of other serializable types
        * Optional types (and types that are equivalent to Optional types) of other
            serializable types

    Users can also supply serializers for custom types by providing a dictionary of type to
    a tuple of serialization and deserialization functions. Serializers can return either a
    string or a list of strings. Lists of strings are represented on the command line as
    ``--arg-name value1 --arg-name value2``. If you're serializing to a list of strings, the
    deserializer should be prepared to handle either a single string or a list of strings.

    Note:
        While lists, tuples and sets (and their ``Optional`` counterparts) can currently
        serialize empty collections, we can't currently serialize custom types as empty lists.

    """

    def __init__(
        self,
        task_function: Callable,
        serializers: Dict,
        tool_name: str,
        naming_args: Optional[List[str]] = None,
        max_attempts: Optional[int] = None,
        module_source_path: Optional[str] = None,
        name_func: Optional[Callable] = None,
        default_cluster_name: str = "",
        default_compute_resources: Optional[Dict[str, Any]] = None,
        default_resource_scales: Optional[Dict[str, float]] = None,
        yaml_file: Optional[str] = None,
    ) -> None:
        """Initialize TaskGenerator.

        Generates and saves the task template.

        Args:
            task_function: The function that the task will run.
            serializers: A dict mapping types to two callables. The first callable is the
                serializer, which converts an object of the type to a string. And the second is
                a deserializer, which converts a string to an object of the type.
            tool_name: A jobmon tool name for generating tasks.
            naming_args: A list of arguments to use in the task name. If not provided (or
                ``None``), it uses all the arguments. If ``[]`` is provided, it uses no naming
                arguments, and the name of the task will just be the name of the task function.
            max_attempts: The max number of attempts jobmon will make on the tasks
            max_attempts: The max number of attempts jobmon will make on the tasks
            module_source_path: The path to the module source code. If not provided,
                                the module is assumed to be installed in the system.
            name_func: A function that takes in the task name prefix and the kwargs to generate
                the task name. If not provided, the FHS default is used.
            default_cluster_name: The default cluster name to use when creating tasks. If not
                provided, an empty string is used.
            default_compute_resources: The default compute resources when creating tasks.
                If not provided, an empty dictionary is used.
            default_resource_scales: The default resource scales to use when creating tasks.
                If not provided, an empty dictionary is used.
            yaml_file: The path to the yaml file that contains the task template. If not
                provided, the default template is used.
        """
        self.task_function = task_function
        self.serializers = serializers
        self.tool = Tool(tool_name)
        self.max_attempts = max_attempts
        self.mod_name = f"{task_function.__module__}"
        self.name = task_function.__name__
        self.full_path = f"{task_function.__module__}:{self.name}"
        self.module_source_path = module_source_path
        self.params = {
            name: details.annotation
            for name, details in inspect.signature(
                self.task_function
            ).parameters.items()
        }
        if naming_args is None:
            logger.warning(
                "You have specified no naming_args on a task_generator, which means that all "
                "task args will be used in the task name. Usually, this results in a name "
                "that's too long for Jobmon, so you might need to specify naming_args. "
                "Include just the arguments that differ from task to task within a workflow."
            )
        self._naming_args = (
            naming_args if naming_args is not None else self.params.keys()
        )

        self._validate_task_function()

        self._task_template = None
        self._default_cluster_name = default_cluster_name
        self._default_compute_resources = default_compute_resources
        self._default_resource_scales = default_resource_scales
        self._yaml_file = yaml_file

        # If the user provides a name_func, use that to generate the task name
        # otherwise, use FHS default
        self.name_func = name_func

    def _validate_task_function(self) -> None:
        """Check that a task can be generated from the task_function.

        Currently checks:
        - All parameters have type annotations
        - All parameters have serializers
        """
        # Verify that all parameters have type annotations
        unannotated_params = [
            name
            for name, annotation in self.params.items()
            if annotation == inspect._empty
        ]
        if unannotated_params:
            raise TypeError(
                f"Can't generate a task template for {self.name} because "
                "the following parameters are missing type annotations: "
                f"{' '.join(unannotated_params)}"
            )

        # Verify that we know how to serialize all of the parameters
        for arg_name, annotation in self.params.items():
            if not self._is_valid_annotation(annotation):
                raise TypeError(
                    f"Unknown annotation on task_function ``{self.name}`` for parameter "
                    f"``{arg_name}: {annotation}``. Please either provide a serializer and "
                    f"deserializer for this type in the ``serializers`` dictionary, or "
                    f"annotate this parameter with one of the known types: "
                    f"{SIMPLE_TYPES | BUILT_IN_COLLECTIONS | OPTIONAL_TYPES}."
                )

    def _is_valid_annotation(self, annotation: Type) -> bool:
        """Returns True if the annotation is valid, False otherwise."""
        # Check for simple types and types that were provided in the serializers dict
        if annotation in SIMPLE_TYPES:
            return True

        if annotation in self.serializers.keys():
            return True

        # Check for built-in collection types. Note that we validate the underlying type
        if _is_annotated_built_in_collection_type(annotation):
            if all(
                # The type can either be a simple type, a type in the serializers dict, or
                # the Ellipses (``...``) type
                t in SIMPLE_TYPES | set(self.serializers.keys()) | {Ellipsis}
                for t in _get_generic_type_parameters(annotation)
            ):
                return True

        # Check for optional types. Note that we validate the underlying type
        if is_optional_type(annotation):
            return self._is_valid_annotation(get_optional_type_parameter(annotation))

        # If we've made it through all the conditions above, return False
        return False

    def _generate_task_template(self) -> None:
        """Generate and store the task template."""
        # args convert to --args foo=1 --args bar=2
        args_template = " --args ".join(
            f"{arg_name}={{{arg_name}}}" for arg_name in self.params
        )
        args_template = " --args " + args_template
        if self.module_source_path:
            self._task_template = self.tool.get_task_template(
                template_name=self.name,
                command_template="{executable} "
                + TASK_RUNNER_SUB_COMMAND
                + " --module_name "
                + self.mod_name
                + " --func_name "
                + self.name
                + " --module_source_path "
                + self.module_source_path
                + args_template,
                node_args=self.params.keys(),  # type: ignore
                op_args=["executable"],
                default_cluster_name=self._default_cluster_name,
                default_compute_resources=self._default_compute_resources,
                default_resource_scales=self._default_resource_scales,
                yaml_file=self._yaml_file,
            )
        else:
            self._task_template = self.tool.get_task_template(
                template_name=self.name,
                command_template="{executable} "
                + TASK_RUNNER_SUB_COMMAND
                + " --module_name "
                + self.mod_name
                + " --func_name "
                + self.name
                + args_template,
                node_args=self.params.keys(),  # type: ignore
                op_args=["executable"],
                default_cluster_name=self._default_cluster_name,
                default_compute_resources=self._default_compute_resources,
                default_resource_scales=self._default_resource_scales,
                yaml_file=self._yaml_file,
            )

    def _is_valid_type(self, obj: Any, expected_type: Type) -> bool:
        """Check that the obj type matches the expected type."""
        if _get_collection_type(expected_type) in BUILT_IN_COLLECTIONS:
            matches = isinstance(obj, _get_collection_type(expected_type))  # type: ignore
        elif is_optional_type(expected_type):
            if obj is None:
                matches = True
            else:
                matches = self._is_valid_type(
                    obj, get_optional_type_parameter(expected_type)
                )
        else:
            matches = isinstance(obj, expected_type)

        return matches

    def serialize(self, obj: Any, expected_type: Type) -> str:
        """Serialize ``obj``, validating that it is actually of type ``expected_type``."""
        if not self._is_valid_type(obj, expected_type):
            raise TypeError(
                f"Expected something of type ``{expected_type}``, but got ``{obj}`` of "
                f"type ``{type(obj)}``."
            )

        serialized_result: str

        # raise an error if the expected_type is a multi-dimensional collection
        if _is_multidimensional_type(expected_type):
            raise TypeError(
                f"This version of Task Generator cannot "
                f"serialize multi-dimensional collection: {expected_type}."
            )

        if expected_type in self.serializers.keys():
            # The 0'th index of the serializers dict is the serialization function
            serialized_result = self.serializers[expected_type][0](obj)

        elif expected_type in SIMPLE_TYPES:
            # If the object is an empty string use the placeholder, otherwise just stringify it
            if expected_type == str and len(obj) == 0:
                serialized_result = SERIALIZED_EMPTY_STRING
            else:
                serialized_result = str(obj)

        elif is_optional_type(expected_type):
            if obj is None:
                serialized_result = str(None)
            else:
                serialized_result = self.serialize(
                    obj=obj, expected_type=get_optional_type_parameter(expected_type)
                )

        elif _get_collection_type(expected_type) in BUILT_IN_COLLECTIONS:

            internal_type = _get_generic_type_parameters(expected_type)
            if internal_type is None:
                raise TypeError(
                    f"Cannot serialize collection with unknown internal type: {expected_type}."
                )

            serialized_result = [
                self.serialize(obj=item, expected_type=item_type)  # type: ignore
                for item, item_type in _zip_collection_items_and_types(
                    obj, internal_type
                )
            ]
            # convert list to string
            if isinstance(serialized_result, list):
                serialized_result = f"[{','.join(serialized_result)}]"

        else:
            raise TypeError(
                f"Cannot serialize unknown type {expected_type.__name__}. Upon "
                "constructing this TaskGenerator, please provide serialization and "
                "deserialization functions for this type in the ``serializers`` dictionary."
            )

        # error out if the output is not a string
        if not isinstance(serialized_result, str):
            raise TypeError(
                f"Expected a string to serialize, but got {type(serialized_result)}."
            )
        return serialized_result

    def serialize_array(self, obj: Any, expected_type: Type) -> List:
        """Serialize obj into a list of serialized string."""
        if isinstance(obj, list):
            return [self.serialize(item, expected_type) for item in obj]
        else:
            return [self.serialize(obj, expected_type)]

    def deserialize(self, obj: str, obj_type: Type) -> Any:
        """Deserialize ``obj``."""
        deserialized_result: Any
        # error out if the input is not a string
        if not isinstance(obj, str):
            raise TypeError(f"Expected a string to deserialize, but got {type(obj)}.")

        # raise an error if the expected_type is a multi-dimensional collection
        if _is_multidimensional_type(obj_type):
            raise TypeError(
                f"This version of Task Generator cannot "
                f"serialize multi-dimensional collection: {obj_type}."
            )

        if is_optional_type(obj_type):
            if obj == str(None):
                return None
            else:
                obj_type = get_optional_type_parameter(obj_type)

        if obj_type in self.serializers.keys():
            # To support dynamic length list input, the deserializer only takes string
            if not isinstance(obj, str):
                raise TypeError(
                    f"Expected a string to deserialize, but got {type(obj)}."
                )
            # The 1'st index of the serializers dict is the deserialization function
            deserialized_result = self.serializers[obj_type][1](obj)

        elif obj_type in SIMPLE_TYPES:
            # If we simply called something like ``bool("False")`` we would get ``True`` due to
            # the way Python thinks about truthiness. So instead we use ``ast.literal_eval``
            # to evaluate the string as a Python literal.
            if obj_type == bool:
                deserialized_result = ast.literal_eval(obj)  # type: ignore

            # If the object is our serialized empty string, return an empty string
            elif obj_type == str and obj == SERIALIZED_EMPTY_STRING:
                deserialized_result = ""

            # For all other simple types, we can simply call the type on ``obj``
            else:
                deserialized_result = obj_type(obj)

        elif _is_unannotated_built_in_collection_type(obj_type):
            raise TypeError(
                f"The provided type annotation ``{obj_type}`` does not provide enough "
                "information to deserialize.  Please provide a more specific type "
                "annotation like ``list[int]`` or ``tuple[str]``."
            )

        elif _is_annotated_built_in_collection_type(obj_type):
            # input is a string, convert it to a list
            try:
                middle_result = ast.literal_eval(obj)
                # if the input is a single item, convert it to a list
                if type(middle_result) not in BUILT_IN_COLLECTIONS:
                    middle_result = [middle_result]
            except Exception:
                # handle input like "[a,b,c]"
                # remove leading and tailing space
                obj = obj.strip()
                # handle input like "'[a,b,c]'"
                if obj[0] == "'" and obj[-1] == "'":
                    obj = obj[1:-1]
                # remove leading and tailing brackets
                if obj[0] in ["[", "("] and obj[-1] in ["]", ")"]:
                    obj = obj[1:-1]

                middle_result = obj.split(",")

            deserialized_result = []
            for item in middle_result:
                if obj_type.__args__[0] in self.serializers.keys():
                    deserialized_result.append(
                        self.serializers[obj_type.__args__[0]][1](item)
                    )
                else:
                    deserialized_result.append(obj_type.__args__[0](item))

        else:
            raise TypeError(
                f"Cannot deserialize unknown type {obj_type.__name__}. Upon "
                "constructing this TaskGenerator, please provide serialization and "
                "deserialization functions for this type in the ``serializers`` dictionary."
            )

        return deserialized_result

    def _cluster_resource_check(
        self, cluster_name: str, compute_resources: Optional[Dict]
    ) -> str:
        """Make sure a cluster name is available to use, and compute_resource is a dict."""
        # add compute_resources type protection
        if compute_resources:
            if not isinstance(compute_resources, dict):
                raise TypeError(
                    f"Expected a dictionary for compute_resources, "
                    f"but got {type(compute_resources)}."
                )
        # if cluster_name is still an empty string, use the default cluster name
        if cluster_name == "":
            cluster_name = self._default_cluster_name
        # if cluster_name is still an empty string, leave to wf to handle
        logger.info(
            "Cluster name is empty. Will use the wf default if one is provided."
        )
        return cluster_name

    def create_task(
        self,
        cluster_name: str = "",
        compute_resources: Optional[Dict] = None,
        resource_scales: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> Task:
        """Create a task for the task_function with the given kwargs."""
        # if the task template is not generated, generate it
        if self._task_template is None:
            self._generate_task_template()

        # check input and adjust the cluster_name value
        cluster_name = self._cluster_resource_check(cluster_name, compute_resources)

        executable_path = _find_executable_path(executable_name=TASK_RUNNER_NAME)
        # Serialize the kwargs
        serialized_kwargs = {
            name: self.serialize(obj=value, expected_type=self.params[name])
            for name, value in kwargs.items()
        }

        # assign the None value to args in self.params.keys but not in serialized_kwargs.keys
        for name in self.params.keys():
            if name not in serialized_kwargs.keys():
                serialized_kwargs[name] = str(None)

        # Format the kwargs for the task
        kwargs_for_task = {
            name: make_cli_argument_string(arg_value=value)
            for name, value in serialized_kwargs.items()
        }
        # We want a slightly different format to put list kwargs into the name
        kwargs_for_name = {
            name: value
            for name, value in serialized_kwargs.items()
            if name in self._naming_args
        }

        name = create_task_name(
            kwargs_for_name=kwargs_for_name, prefix=self.name, name_func=self.name_func
        )

        # Create the task
        task = self._task_template.create_task(  # type: ignore
            name=name,
            cluster_name=cluster_name,
            compute_resources=compute_resources,
            resource_scales=resource_scales,
            max_attempts=self.max_attempts,
            executable=executable_path,
            **kwargs_for_task,  # type: ignore
        )

        return task

    def create_tasks(
        self,
        cluster_name: str = "",
        compute_resources: Optional[Dict] = None,
        resource_scales: Optional[Dict[str, Any]] = None,
        **kwargs: Any,
    ) -> List[Task]:
        """Create a task array for the task_function with the given kwargs."""
        # if the task template is not generated, generate it
        if self._task_template is None:
            self._generate_task_template()

        # check input and adjust the cluster_name value
        cluster_name = self._cluster_resource_check(cluster_name, compute_resources)

        executable_path = _find_executable_path(executable_name=TASK_RUNNER_NAME)
        # Serialize the kwargs
        serialized_kwargs = {
            name: self.serialize_array(obj=value, expected_type=self.params[name])
            for name, value in kwargs.items()
        }

        # Format the kwargs for the task
        # Each individual element in the array should be formatted before sending to the cli
        kwargs_for_task = {}
        for name, value in serialized_kwargs.items():
            if isinstance(value, list):
                kwargs_for_task[name] = [
                    make_cli_argument_string(arg_value=item) for item in value
                ]
            else:
                kwargs_for_task[name] = make_cli_argument_string(arg_value=value)

        # name is auto for array

        # Create the task
        tasks = self._task_template.create_tasks(  # type: ignore
            cluster_name=cluster_name,
            compute_resources=compute_resources,
            resource_scales=resource_scales,
            max_attempts=self.max_attempts,
            executable=executable_path,
            **kwargs_for_task,  # type: ignore
        )

        return tasks

    def run(self, args: List[str]) -> Any:
        """Run the task_function with the given args and return any result."""
        # Parse the args
        parsed_arg_value_pairs: Dict[str, Union[str, List[str]]] = dict()
        # args is a list of string like ["arg1=1", "arg2=[2, 3]", "arg1=4]
        for arg in args:
            arg_name, arg_value = arg.split("=")
            # if the arg_name key, already exists, append the value to the list
            if arg_name in parsed_arg_value_pairs.keys():
                if isinstance(parsed_arg_value_pairs[arg_name], list):
                    parsed_arg_value_pairs[arg_name].append(arg_value)  # type: ignore
                else:
                    parsed_arg_value_pairs[arg_name] = [
                        parsed_arg_value_pairs[arg_name],  # type: ignore
                        arg_value,
                    ]
            else:
                parsed_arg_value_pairs[arg_name] = arg_value  # type: ignore

        # Raise an error if the user did not provide all of the arguments for the task_function
        if parsed_arg_value_pairs.keys() != self.params.keys():
            raise ValueError(
                "Some arguments to the task_function were not provided on the command line. "
                "The following arguments were not provided: "
                f"{set(self.params.keys()) - set(parsed_arg_value_pairs.keys())}"
            )

        # Deserialize the args, catching any errors that may come from the deserialization fn
        deserialized_args = dict()
        for arg_name, arg_value in parsed_arg_value_pairs.items():  # type: ignore
            try:
                deserialized_args[arg_name] = self.deserialize(
                    obj=arg_value, obj_type=self.params[arg_name]
                )
            except Exception:
                raise ValueError(
                    f"Could not deserialize argument ``{arg_name}`` with value "
                    f"``{arg_value}``. This error could be the result of a bug in the "
                    f"deserialization function, or a result of not providing a value for "
                    f"this argument on the command line."
                )

        # Run the task_function
        return self.task_function(**deserialized_args)

    def help(self) -> str:
        """Return help text for the task_function."""
        # Parse the task function's docstring - Note that there may be nothing!
        task_function_docstring = docstring_parser.parse(  # type: ignore
            self.task_function.__doc__  # type: ignore
        )  # type: ignore

        # Map the parameter names to their annotations and descriptions
        task_function_docstring_param_names_to_annotations = {
            param.arg_name: param.type_name for param in task_function_docstring.params
        }
        task_function_docstring_param_names_to_descriptions = {
            param.arg_name: param.description
            for param in task_function_docstring.params
        }

        # Get the short description
        short_description = _get_short_description(task_function_docstring)

        # Get the parameter names and descriptions
        parameter_names_and_descriptions = self._get_param_names_and_descriptions(
            task_function_docstring_param_names_to_annotations,
            task_function_docstring_param_names_to_descriptions,
        )

        # Combine the full help text and return it
        return "\n".join(
            [
                HELP_TEXT_INTRO_FORMAT.format(full_path=self.full_path),
                short_description,
                *parameter_names_and_descriptions,
            ]
        )

    def _get_param_names_and_descriptions(
        self,
        task_function_docstring_param_names_to_annotations: Dict,
        task_function_docstring_param_names_to_descriptions: Dict,
    ) -> List[str]:
        parameter_names_and_descriptions = []

        for param, annotation in self.params.items():
            # If the parameter has an annotation, use that for the text.
            if param in task_function_docstring_param_names_to_annotations.keys():
                annotation_string = task_function_docstring_param_names_to_annotations[
                    param
                ]

            else:
                # Otherwise try to pull the annotation from the self.params annotation (this
                # will work for simple types like str, int, etc.)
                try:
                    annotation_string = annotation.__name__

                # Otherwise just stringify the annotation (this will work for custom or union
                # types like ``str | None``)
                except AttributeError:
                    annotation_string = str(annotation)

            this_param_text = f"\t--{param} [{annotation_string}]: "

            # If the parameter has a description in the docstring, add that to the text
            if param in task_function_docstring_param_names_to_descriptions.keys():
                this_param_text += (
                    f"{task_function_docstring_param_names_to_descriptions[param]}"
                )
            else:
                this_param_text += "No description found in docstring."

            parameter_names_and_descriptions.append(this_param_text)

        return parameter_names_and_descriptions


def task_generator(
    serializers: Dict,
    tool_name: str,
    naming_args: Optional[List[str]] = None,
    max_attempts: Optional[int] = None,
    module_source_path: Optional[str] = None,
    default_cluster_name: str = "",
    default_compute_resources: Optional[Dict[str, Any]] = None,
    default_resource_scales: Optional[Dict[str, float]] = None,
    yaml_file: Optional[str] = None,
) -> Callable:
    """Decorator for generating jobmon tasks from a python function."""

    def wrapper(task_function: Callable) -> TaskGenerator:
        """Wrap the task_function with a TaskGenerator."""
        return TaskGenerator(
            task_function=task_function,
            serializers=serializers,
            tool_name=tool_name,
            naming_args=naming_args,
            max_attempts=max_attempts,
            module_source_path=module_source_path,
            default_cluster_name=default_cluster_name,
            default_compute_resources=default_compute_resources,
            default_resource_scales=default_resource_scales,
            yaml_file=yaml_file,
        )

    return wrapper


def get_tasks_by_node_args(
    workflow: Workflow,
    task_generator: TaskGenerator,
    node_args_dict: dict[str, Any],
    error_on_empty: bool = True,
) -> list[Task]:
    """Get the tasks of a TaskGenerator in a workflow that have the given node arguments.

    This method does some value serialization and formatting before handing the node argument
    string over to ``workflow.get_tasks_by_node_args``.
    """
    # Re-serialize the node args dict and format them as CLI arguments
    serialized_node_args = {
        arg_name: make_cli_argument_string(
            arg_value=task_generator.serialize(
                obj=arg_value, expected_type=task_generator.params[arg_name]
            ),
        )
        for arg_name, arg_value in node_args_dict.items()
    }

    try:
        result = workflow.get_tasks_by_node_args(
            task_template_name=task_generator.name, **serialized_node_args
        )
    except ValueError as err:
        if error_on_empty:
            raise err
        result = []

    if not result and error_on_empty:
        raise ValueError(
            f"There were no tasks in the workflow that matched {node_args_dict}"
        )

    return result


class TaskGeneratorDocumenter(Directive):
    """Directive for generating documentation for a single task generator."""

    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {"optional": lambda x: x}  # Defines the 'optional' option

    def run(self) -> list[nodes.Node]:
        """The function sphinx/docutils use to generate documentation for the directive."""
        module_name = self.arguments[0]
        # if there are more than one arg, the second will be module path
        module_path = self.options.get("optional", None)
        task_generator = self._load_task_generator(module_name, module_path)

        return _generate_nodes(task_generator, self.state)

    def _load_task_generator(
        self, task_generator_path: str, module_path: Optional[str] = None
    ) -> TaskGenerator:
        """Load the task generator from the given path."""
        try:
            module_name, attr_name = task_generator_path.split(":", 1)
        except ValueError:
            raise self.error(
                f'"{task_generator_path}" is not of format "module:parser"'
            )

        try:
            if module_path:
                spec = importlib.util.spec_from_file_location(
                    module_name, module_path
                )  # type: ignore
                mod = importlib.util.module_from_spec(spec)  # type: ignore
                spec.loader.exec_module(mod)  # type: ignore
            else:
                mod = __import__(module_name, globals(), locals(), [attr_name])
        except (Exception, SystemExit) as exc:
            err_msg = f'Failed to import "{attr_name}" from "{module_name}". '
            if isinstance(exc, SystemExit):
                err_msg += "The module appeared to call sys.exit()"
            else:
                err_msg += (
                    f"The following exception was raised:\n{traceback.format_exc()}"
                )
            raise self.error(err_msg)

        if not hasattr(mod, attr_name):
            raise self.error(f'Module "{module_name}" has no attribute "{attr_name}"')

        task_generator = getattr(mod, attr_name)

        if not isinstance(task_generator, TaskGenerator):
            raise self.error(
                f'Attribute "{attr_name}" of module "{module_name}" is not a TaskGenerator'
            )

        return task_generator


class TaskGeneratorModuleDocumenter(Directive):
    """Directive for generating documentation for all the task generators in a module."""

    required_arguments = 1
    optional_arguments = 1
    final_argument_whitespace = True
    option_spec = {"optional": lambda x: x}  # Defines the 'optional' option

    def run(self) -> list[nodes.Node]:
        """The function sphinx/docutils use to generate documentation for the directive."""
        module_name = self.arguments[0]
        # if there are more than one arg, the second will be module path
        module_path = self.options.get("optional", None)
        module, task_generators = self._load_module_task_generators(
            module_name, module_path
        )

        section = self._generate_module_section(module_name, module, task_generators)

        for task_generator in task_generators:
            section.extend(_generate_nodes(task_generator, self.state))

        return [section]

    def _load_module_task_generators(
        self, module_name: str, module_path: Optional[str] = None
    ) -> tuple[ModuleType, list[TaskGenerator]]:
        """Load all the task generators in a given module."""
        task_generators = []

        try:
            if module_path:
                spec = importlib.util.spec_from_file_location(
                    module_name, module_path
                )  # type: ignore
                mod = importlib.util.module_from_spec(spec)  # type: ignore
                spec.loader.exec_module(mod)  # type: ignore
            else:
                mod = importlib.import_module(module_name)
        except (Exception, SystemExit) as exc:
            err_msg = f'Failed to import "{module_name}". '
            if isinstance(exc, SystemExit):
                err_msg += "The module appeared to call sys.exit()"
            else:
                err_msg += (
                    f"The following exception was raised:\n{traceback.format_exc()}"
                )
            raise self.error(err_msg)

        for attr_name in dir(mod):
            if attr_name.startswith("_"):
                continue

            attr = getattr(mod, attr_name)

            if isinstance(attr, TaskGenerator):
                task_generators.append(attr)

        return mod, task_generators

    def _generate_module_section(
        self,
        module_name: str,
        module: ModuleType,
        task_generators: Iterable[TaskGenerator],
    ) -> nodes.Node:
        """Makes the docutils node for the module section.

        Includes the docstring for the module, and a list of task generator names. Doesn't
        include any task generator detailed documentation.
        """
        # Set up base node
        section = nodes.section(
            "",
            nodes.title(text=module_name),
            ids=[nodes.make_id(module_name)],
            names=[nodes.fully_normalize_name(module_name)],
        )

        # Use sphinx and docutils tooling to format the docstring into rST and then parse it
        # into a docutils node.
        result = statemachine.ViewList()
        if module.__doc__:
            for line in statemachine.string2lines(
                module.__doc__, tab_width=4, convert_whitespace=True
            ):
                result.append(line, module_name)

        task_generator_path_lines = [""]
        for task_generator in task_generators:
            task_generator_path_lines.append(".. code-block:: shell")
            task_generator_path_lines.append("")
            task_generator_path_lines.append(
                f"    {TASK_RUNNER_NAME} {task_generator.full_path}"
            )
            task_generator_path_lines.append("")

        for line in task_generator_path_lines:
            result.append(line, module_name)

        sphinx_nodes.nested_parse_with_titles(self.state, result, section)

        return section


def _generate_nodes(task_generator: TaskGenerator, state: Any) -> nodes.Node:
    """Makes a docutils node with the documentation for a task generator.

    Includes the docstring description for the task generator and all the arguments.
    """
    # Set up base node
    section = nodes.section(
        "",
        nodes.title(text=task_generator.name),
        ids=[nodes.make_id(task_generator.full_path)],
        names=[nodes.fully_normalize_name(task_generator.full_path)],
    )

    # Get rST lines for the description and options
    parsed_docstring = docstring_parser.parse(str(task_generator.task_function.__doc__))
    lines = []
    lines.extend(
        _format_description(
            task_generator=task_generator, parsed_docstring=parsed_docstring
        )
    )
    lines.extend(
        _format_options(
            task_generator=task_generator, parsed_docstring=parsed_docstring
        )
    )

    # Convert the rST lines into a docutils node
    result = statemachine.ViewList()
    for line in lines:
        result.append(line, task_generator.name)
    sphinx_nodes.nested_parse_with_titles(state, result, section)

    return [section]


def _format_description(
    task_generator: TaskGenerator, parsed_docstring: docstring_parser.Docstring
) -> list[str]:
    """Format the description of the task generator into proper rST."""
    lines = []

    # Description from the docstring
    if parsed_docstring.short_description:
        lines.append(parsed_docstring.short_description)
        lines.append("")

    if parsed_docstring.long_description:
        for line in statemachine.string2lines(
            parsed_docstring.long_description, tab_width=4, convert_whitespace=True
        ):
            lines.append(line)
        lines.append("")

    # How to run on the cli
    lines.append(".. code-block:: shell")
    lines.append("")
    lines.append(f"    {TASK_RUNNER_NAME} {task_generator.full_path} [OPTIONS]")
    lines.append("")

    return lines


def _format_options(
    task_generator: TaskGenerator, parsed_docstring: docstring_parser.Docstring
) -> list[str]:
    """Format the options of the task generator into proper rST."""
    param_docs = {param.arg_name: param for param in parsed_docstring.params}
    lines = []

    # we use rubric to provide some separation without exploding the table
    # of contents
    lines.append(".. rubric:: Options")
    lines.append("")
    for param, annotation in task_generator.params.items():
        if is_optional_type(annotation):
            underlying_optional_type = get_optional_type_parameter(annotation)
            annotation_name = f"OPTIONAL[{underlying_optional_type.__name__.upper()}]"
        elif isinstance(annotation, type):
            annotation_name = annotation.__name__.upper()
        else:
            # it can be typing._GenericAlias for list type annotation
            annotation_name = str(annotation).upper()

        lines.append(f".. option:: --{param} <{annotation_name}>")
        lines.append("")
        if param in param_docs and param_docs[param].description:
            for line in statemachine.string2lines(
                param_docs[param].description, tab_width=4, convert_whitespace=True
            ):
                lines.append(line)
            lines.append("")

    return lines


def setup(app: application.Sphinx) -> dict:
    """The function that registers the extension with sphinx."""
    app.add_directive("task_generator", TaskGeneratorDocumenter)
    app.add_directive("task_generator_module", TaskGeneratorModuleDocumenter)

    return {
        "version": "0.1",
        "parallel_read_safe": True,
        "parallel_write_safe": True,
    }



"""Command line interface for Execution."""

import argparse
from typing import Optional

from jobmon.core.cli import CLI
from jobmon.core.cluster import Cluster
from jobmon.distributor.api import DistributorService


class DistributorCLI(CLI):
    """Command line interface for Distributor."""

    def __init__(self) -> None:
        """Initialization of distributor CLI."""
        self.parser = argparse.ArgumentParser()
        self._subparsers = self.parser.add_subparsers(
            dest="sub_command", parser_class=argparse.ArgumentParser
        )

        self._add_distributor_parser()

    @staticmethod
    def run_distributor(args: argparse.Namespace) -> None:
        """Configuration for the jobmon worker node."""
        cluster = Cluster.get_cluster(args.cluster_name)
        cluster_interface = cluster.get_distributor()
        distributor_service = DistributorService(cluster_interface)
        distributor_service.set_workflow_run(args.workflow_run_id)
        distributor_service.run()

    def _add_distributor_parser(self) -> None:
        distributor_parser = self._subparsers.add_parser("start")
        distributor_parser.set_defaults(func=self.run_distributor)
        distributor_parser.add_argument(
            "--cluster_name",
            type=str,
            help="cluster_name to distribute jobs onto.",
            required=True,
        )
        distributor_parser.add_argument(
            "--workflow_run_id",
            type=int,
            help="workflow_run_id to distribute jobs for.",
            required=True,
        )


def main(argstr: Optional[str] = None) -> None:
    """Entrypoint to create Executor CLI."""
    cli = DistributorCLI()
    cli.main(argstr)


if __name__ == "__main__":
    main()



"""Dummy Executor fakes execution for testing purposes."""

from __future__ import annotations

from importlib.metadata import version
import logging
from typing import Type

from jobmon.core.cluster_protocol import (
    ClusterDistributor,
    ClusterQueue,
    ClusterWorkerNode,
)

logger = logging.getLogger(__name__)

__version__ = version("jobmon_core")


def get_cluster_queue_class() -> Type[ClusterQueue]:
    """Return the queue class for the dummy executor."""
    from jobmon.plugins.dummy.dummy_queue import DummyQueue

    return DummyQueue


def get_cluster_distributor_class() -> Type[ClusterDistributor]:
    """Return the cluster distributor for the dummy executor."""
    from jobmon.plugins.dummy.dummy_distributor import DummyDistributor

    return DummyDistributor


def get_cluster_worker_node_class() -> Type[ClusterWorkerNode]:
    """Return the cluster worker node class for the dummy executor."""
    from jobmon.plugins.dummy.dummy_distributor import DummyWorkerNode

    return DummyWorkerNode



from __future__ import annotations

import getpass
import logging
import os
import socket
import sys
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Type, Union


from opentelemetry import _logs, trace
from opentelemetry.sdk import resources
from opentelemetry.sdk._logs import LoggerProvider
from opentelemetry.sdk._logs.export import BatchLogRecordProcessor
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.trace import Tracer

from jobmon.core import __version__
from jobmon.core.configuration import JobmonConfig


class OtlpAPI:
    """OpenTelemetry API."""

    _instance = None
    _initialized = False
    _sqlalchemy_instrumented = False
    _requests_instrumented = False

    def __new__(cls: Type[OtlpAPI], *args: Any, **kwargs: Any) -> OtlpAPI:
        if not cls._instance:
            cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(
        self, extra_detectors: Optional[List[resources.ResourceDetector]] = None
    ) -> None:
        """Initialize the OtlpAPI object."""
        if OtlpAPI._initialized:
            return

        if extra_detectors is None:
            extra_detectors = []

        self._configure_resources(extra_detectors)
        self._configure_providers()

        OtlpAPI._initialized = True

    def _configure_resources(
        self, extra_detectors: List[resources.ResourceDetector]
    ) -> None:
        self._detectors = [
            _ProcessResourceDetector(),
            _HostResourceDetector(),
            _ServiceResourceDetector(),
        ]

        if extra_detectors:
            self._detectors.extend(extra_detectors)

        resource_group = resources.get_aggregated_resources(self._detectors)

        # Store the SDK TracerProvider instance
        self.tracer_provider = TracerProvider(resource=resource_group)
        trace.set_tracer_provider(self.tracer_provider)

        # Store the SDK LoggerProvider instance
        self.logger_provider = LoggerProvider(resource=resource_group)
        _logs.set_logger_provider(self.logger_provider)

    def _configure_providers(self) -> None:
        config = JobmonConfig()

        span_exporter = config.get("otlp", "span_exporter")
        if span_exporter:
            span_kwargs = config.get_section(span_exporter)
            self._set_exporter(
                span_kwargs,
                self.tracer_provider.add_span_processor,  # Use the SDK instance
                BatchSpanProcessor,
            )

        log_exporter = config.get("otlp", "log_exporter")
        if log_exporter:
            log_kwargs = config.get_section(log_exporter)
            self._set_exporter(
                log_kwargs,
                self.logger_provider.add_log_record_processor,  # Use the SDK instance
                BatchLogRecordProcessor,
            )

    def _set_exporter(
        self,
        kwargs: Any,
        add_processor_func: Callable[[Any], None],
        batch_processor_class: Type[Any],
    ) -> None:
        module_name = kwargs["module"]
        class_name = kwargs["class"]
        module = __import__(module_name, fromlist=[class_name])
        ExporterClass = getattr(module, class_name)
        processor = batch_processor_class(
            ExporterClass(
                **{k: v for k, v in kwargs.items() if k not in ["module", "class"]}
            )
        )
        add_processor_func(processor)

    @classmethod
    def instrument_sqlalchemy(cls: Type[OtlpAPI]) -> None:
        """Instrument SQLAlchemy."""
        if not cls._sqlalchemy_instrumented:
            from opentelemetry.instrumentation.sqlalchemy import SQLAlchemyInstrumentor

            SQLAlchemyInstrumentor().instrument()
            cls._sqlalchemy_instrumented = True

    @classmethod
    def instrument_app(cls: Type[OtlpAPI], app: Any) -> None:
        """Instrument FastAPI app."""
        from opentelemetry.instrumentation.fastapi import FastAPIInstrumentor

        # Instrument FastAPI with OpenTelemetry
        FastAPIInstrumentor().instrument_app(app)

    @classmethod
    def instrument_requests(cls: Type[OtlpAPI]) -> None:
        """Instrument requests."""
        if not cls._requests_instrumented:
            from opentelemetry.instrumentation.requests import RequestsInstrumentor

            RequestsInstrumentor().instrument()
            cls._requests_instrumented = True

    def get_tracer(self, name: str) -> Tracer:
        """Get a tracer from the SDK TracerProvider."""
        return self.tracer_provider.get_tracer(name)

    def get_logger_provider(self) -> LoggerProvider:
        """Get the logger provider."""
        return self.logger_provider

    @staticmethod
    def get_span_details() -> Tuple[Optional[str], Optional[str], Optional[str]]:
        """Retrieve details of the current span."""
        span = trace.get_current_span()

        # Check if there's a valid span
        if not span or not span.is_recording():
            return None, None, None

        ctx = span.get_span_context()

        # Get parent span, but handle if it doesn't exist
        parent = getattr(span, "parent", None)

        span_id = format(ctx.span_id, "016x") if ctx and ctx.span_id else None
        trace_id = format(ctx.trace_id, "032x") if ctx and ctx.trace_id else None
        parent_span_id = format(parent.span_id, "016x") if parent else None

        return span_id, trace_id, parent_span_id


class OpenTelemetryLogFormatter(logging.Formatter):
    """Formatter that adds OpenTelemetry spans to log records."""

    def format(self, record: logging.LogRecord) -> str:
        span_id, trace_id, parent_span_id = OtlpAPI.get_span_details()
        record.span_id = span_id
        record.trace_id = trace_id
        record.parent_span_id = parent_span_id
        return super().format(record)


def add_span_details_processor(
    logger: Any, method_name: str, event_dict: Dict[str, Any]
) -> Dict[str, Any]:
    """Structlog processor to add OpenTelemetry span details to log entries.

    Args:
        logger: The logger instance (not used, but required by Structlog processor signature).
        method_name: The logging method name (e.g., "info", "debug").
        event_dict: The event dictionary representing the log entry.

    Returns:
        The modified event dictionary with OpenTelemetry span details added.
    """
    span_id, trace_id, parent_span_id = OtlpAPI.get_span_details()
    if trace_id or span_id or parent_span_id:
        event_dict.update(
            {
                "trace_id": trace_id,
                "span_id": span_id,
                "parent_span_id": parent_span_id,
            }
        )
    return event_dict


class _ProcessResourceDetector(resources.ResourceDetector):
    def detect(self) -> resources.Resource:
        attrs: Mapping[str, Union[str, bool, int, float]] = {
            str(resources.PROCESS_PID): int(os.getpid()),  # Explicit cast to int
            str(resources.PROCESS_RUNTIME_NAME): str(
                sys.implementation.name
            ),  # Explicit cast to str
            str(resources.PROCESS_OWNER): str(
                getpass.getuser()
            ),  # Explicit cast to str
        }
        return resources.Resource(attrs)


class _ServiceResourceDetector(resources.ResourceDetector):
    def detect(self) -> resources.Resource:
        attrs = {
            resources.SERVICE_NAME: "jobmon",
            resources.SERVICE_VERSION: __version__,
        }
        return resources.Resource(attrs)


class _HostResourceDetector(resources.ResourceDetector):
    def detect(self) -> resources.Resource:
        attrs = {resources.HOST_NAME: socket.gethostname()}
        return resources.Resource(attrs)



from __future__ import annotations

import ast
import hashlib
import logging
from typing import Any, Dict, Set, TYPE_CHECKING

from jobmon.core.constants import TaskInstanceStatus
from jobmon.core.requester import Requester

if TYPE_CHECKING:
    from jobmon.distributor.distributor_task_instance import (
        DistributorTaskInstance,
    )


logger = logging.getLogger(__name__)


class TaskInstanceBatch:
    def __init__(
        self,
        array_id: int,
        array_name: str,
        array_batch_num: int,
        task_resources_id: int,
        requester: Requester,
    ) -> None:
        """Initialization of the TaskInstanceBatch object."""
        self.array_id = array_id
        self.array_name = array_name
        self.batch_number = array_batch_num
        self.task_resources_id = task_resources_id
        self.task_instances: Set[DistributorTaskInstance] = set()
        self.requester = requester

    @property
    def submission_name(self) -> str:
        return f"{self.array_name}-{self.batch_number}"

    @property
    def requested_resources(self) -> Dict:
        if not hasattr(self, "_requested_resources"):
            raise AttributeError(
                "Requested Resources cannot be accessed before the array batch is prepared for"
                " launch."
            )
        return self._requested_resources

    def add_task_instance(self, task_instsance: DistributorTaskInstance) -> None:
        self.task_instances.add(task_instsance)
        task_instsance.batch = self

    def load_requested_resources(self) -> None:
        app_route = f"/task_resources/{self.task_resources_id}"
        _, response = self.requester.send_request(
            app_route=app_route, message={}, request_type="post"
        )
        self._requested_resources: Dict[str, Any] = ast.literal_eval(
            str(response["requested_resources"])
        )
        self._requested_resources["queue"] = response["queue_name"]

    def prepare_task_instance_batch_for_launch(self) -> None:
        """Add the current batch number to the current set of registered task instance ids."""
        array_step_id = 0
        for task_instance in sorted(self.task_instances):
            task_instance.array_step_id = array_step_id
            array_step_id += 1

        self.load_requested_resources()

    def set_distributor_ids(self, distributor_id_map: Dict) -> None:
        """Set the distributor_ids on the task instances in the array.

        Args:
            distributor_id_map: map of array_step_id to distributor_id
        """
        for ti in self.task_instances:
            ti.distributor_id = distributor_id_map[ti.array_step_id]

    def transition_to_launched(self, next_report_by: float) -> None:
        """Transition all associated task instances to LAUNCHED state."""
        # Assertion that all bound task instances are indeed instantiated
        for ti in self.task_instances:
            if ti.status != TaskInstanceStatus.INSTANTIATED:
                raise ValueError(
                    f"{ti} is not in INSTANTIATED state, prior to launching."
                )

        app_route = f"/array/{self.array_id}/transition_to_launched"
        data = {
            "batch_number": self.batch_number,
            "next_report_increment": next_report_by,
        }

        self.requester.send_request(
            app_route=app_route, message=data, request_type="post"
        )

        for ti in self.task_instances:
            ti.status = TaskInstanceStatus.LAUNCHED

    def log_distributor_ids(self) -> None:
        """Log the distributor ID in the database for all task instances in the batch."""
        app_route = f"/array/{self.array_id}/log_distributor_id"
        data = {ti.task_instance_id: ti.distributor_id for ti in self.task_instances}
        self.requester.send_request(
            app_route=app_route, message=data, request_type="post"
        )

    def transition_to_killed(self) -> None:
        """Mark all TIs in this batch as killed in the DB (ERROR_FATAL)."""
        # 1) Make an HTTP call to the server if you have a relevant endpoint
        #    that transitions them to a 'killed' or ERROR_FATAL state in bulk.
        app_route = f"/array/{self.array_id}/transition_to_killed"
        data = {"batch_number": self.batch_number}
        self.requester.send_request(
            app_route=app_route, message=data, request_type="post"
        )

    def __hash__(self) -> int:
        """Hash to encompass tool version id, workflow args, tasks and dag."""
        hash_value = hashlib.sha1()
        hash_value.update(str(hash(self.array_id)).encode("utf-8"))
        hash_value.update(str(self.batch_number).encode("utf-8"))
        return int(hash_value.hexdigest(), 16)

    def __eq__(self, other: object) -> bool:
        """Check if the hashes of two tasks are equivalent."""
        if not isinstance(other, TaskInstanceBatch):
            return False
        else:
            return hash(self) == hash(other)

    def __lt__(self, other: TaskInstanceBatch) -> bool:
        """Check if one hash is less than the has of another Task."""
        return hash(self) < hash(other)



import argparse
import shlex
from typing import Any, List, Optional


class CLI:
    """Base CLI."""

    def __init__(self) -> None:
        """Initialize the CLI."""
        self.parser = argparse.ArgumentParser("jobmon CLI")

    def main(self, argstr: Optional[str] = None) -> Any:
        """Parse args."""
        args = self.parse_args(argstr)
        return args.func(args)

    def parse_args(self, argstr: Optional[str] = None) -> argparse.Namespace:
        """Construct a parser, parse either sys.argv (default) or the provided argstr.

        Returns a Namespace. The Namespace should have a 'func' attribute which can be used to
        dispatch to the appropriate downstream function.
        """
        arglist: Optional[List[str]] = None
        if argstr is not None:
            arglist = shlex.split(argstr)
        args = self.parser.parse_args(arglist)
        return args



"""The Task Instance Object once it has been submitted to run on a worker node."""

import asyncio
import logging
import logging.config
import os
import signal
import socket
import sys
from time import time
from typing import Dict, Optional, TextIO

from jobmon.core.cluster_protocol import ClusterWorkerNode
from jobmon.core.configuration import JobmonConfig
from jobmon.core.constants import TaskInstanceStatus
from jobmon.core.exceptions import ReturnCodes, TransitionError
from jobmon.core.requester import Requester
from jobmon.core.serializers import SerializeTaskInstance

logger = logging.getLogger(__name__)


class WorkerNodeTaskInstance:
    """The Task Instance object once it has been submitted to run on a worker node."""

    def __init__(
        self,
        cluster_interface: ClusterWorkerNode,
        task_instance_id: int,
        task_instance_heartbeat_interval: Optional[int] = None,
        heartbeat_report_by_buffer: Optional[float] = None,
        command_interrupt_timeout: Optional[int] = None,
        requester: Optional[Requester] = None,
    ) -> None:
        """A mechanism whereby a running task_instance can communicate back to the JSM.

         Logs its status, errors, usage details, etc.

        Args:
            cluster_interface: interface that gathers executor info in the execution_wrapper.
            task_instance_id: the id of the TaskInstance that is reporting back.
            task_instance_heartbeat_interval: how ofter to log a report by with the db
            heartbeat_report_by_buffer: multiplier for report by date in case we miss a few.
            command_interrupt_timeout: the amount of time to wait for the child process to
                terminate.
            requester: communicate with the flask services.
        """
        # identity attributes
        self._task_instance_id = task_instance_id

        # service API
        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

        # cluster API
        self.cluster_interface = cluster_interface

        # get distributor id from executor
        self._distributor_id = self.cluster_interface.distributor_id

        # config
        config = JobmonConfig()
        if task_instance_heartbeat_interval is None:
            self._task_instance_heartbeat_interval = config.get_int(
                "heartbeat", "task_instance_interval"
            )
        else:
            self._task_instance_heartbeat_interval = task_instance_heartbeat_interval
        if heartbeat_report_by_buffer is None:
            self._heartbeat_report_by_buffer = config.get_float(
                "heartbeat", "report_by_buffer"
            )
        else:
            self._heartbeat_report_by_buffer = heartbeat_report_by_buffer
        if command_interrupt_timeout is None:
            self._command_interrupt_timeout = config.get_int(
                "worker_node", "command_interrupt_timeout"
            )
        else:
            self._command_interrupt_timeout = command_interrupt_timeout

        # attrs set by log running
        self._status: Optional[str] = None
        self._command: Optional[str] = None
        self._command_add_env: Optional[Dict[str, str]] = None
        self._stdout: Optional[str] = None
        self._stderr: Optional[str] = None

        # set last heartbeat
        self.last_heartbeat_time = time()

    @property
    def task_instance_id(self) -> int:
        """Returns a task instance ID if it's been bound."""
        if self._task_instance_id is None:
            raise AttributeError("Cannot access task_instance_id because it is None.")
        return self._task_instance_id

    @property
    def distributor_id(self) -> Optional[str]:
        """Executor id given from the executor it is being run on."""
        return self._distributor_id

    @property
    def nodename(self) -> Optional[str]:
        """Node it is being run on."""
        if not hasattr(self, "_nodename"):
            self._nodename = socket.getfqdn()
        return self._nodename

    @property
    def process_group_id(self) -> Optional[int]:
        """Process group to track parent and child processes."""
        if not hasattr(self, "_process_group_id"):
            self._process_group_id = os.getpid()
        return self._process_group_id

    @property
    def status(self) -> str:
        """Returns the last known status of the task instance."""
        if self._status is None:
            raise AttributeError(
                "Cannot access status until log_running() has been called."
            )
        return self._status

    @property
    def stdout(self) -> str:
        if self._stdout is None:
            raise AttributeError(
                "Cannot access stdout until log_running() has been called."
            )
        return self._stdout

    @property
    def stderr(self) -> str:
        if self._stderr is None:
            raise AttributeError(
                "Cannot access stderr until log_running() has been called."
            )
        return self._stderr

    @property
    def command(self) -> str:
        """Returns the command this task instance will run."""
        if self._command is None:
            raise AttributeError(
                "Cannot access command until log_running() has been called."
            )
        return self._command

    @property
    def command_add_env(self) -> Dict[str, str]:
        """Returns the command this task instance will run."""
        if self._command_add_env is None:
            raise AttributeError(
                "Cannot access command_add_env until log_running() has been called."
            )
        return self._command_add_env

    @property
    def command_returncode(self) -> int:
        """Returns the exit code of the command that was run."""
        if not hasattr(self, "_proc_returncode"):
            raise AttributeError(
                "Cannot access command_returncode until run() has been called"
            )
        return self._proc_returncode

    @property
    def command_stdout(self) -> str:
        """Returns the last 10k characters of the commands stdout."""
        if not hasattr(self, "_proc_stdout"):
            raise AttributeError(
                "Cannot access command_stdout until run() has been called"
            )
        return self._proc_stdout

    @property
    def command_stderr(self) -> str:
        """Returns the last 10k characters of the commands stderr."""
        if not hasattr(self, "_proc_stderr"):
            raise AttributeError(
                "Cannot access command_stderr until run() has been called"
            )
        return self._proc_stderr

    def configure_logging(self) -> None:
        """Setup logging for the worker node. INFO level goes to standard out."""
        _DEFAULT_LOG_FORMAT = (
            "%(asctime)s [%(name)-12s] %(module)s %(levelname)-8s: %(message)s"
        )
        logging_config: Dict = {
            "version": 1,
            "disable_existing_loggers": True,
            "formatters": {
                "default": {
                    "format": _DEFAULT_LOG_FORMAT,
                    "datefmt": "%Y-%m-%d %H:%M:%S",
                }
            },
            "handlers": {
                "default": {
                    "level": "INFO",
                    "class": "logging.StreamHandler",
                    "formatter": "default",
                    "stream": sys.stdout,
                },
            },
            "loggers": {
                "jobmon.worker_node": {
                    "handlers": ["default"],
                    "propagate": False,
                    "level": "INFO",
                },
            },
        }
        logging.config.dictConfig(logging_config)

    def log_done(self) -> None:
        """Tell the JobStateManager that this task_instance is done."""
        logger.info(f"Logging done for task_instance {self.task_instance_id}")

        message = {
            "stdout": self.stdout,
            "stderr": self.stderr,
            "stdout_log": self.command_stdout,
            "stderr_log": self.command_stderr,
            "nodename": self.nodename,
            "distributor_id": self.distributor_id,
        }

        app_route = f"/task_instance/{self.task_instance_id}/log_done"
        _, response = self.requester.send_request(
            app_route=app_route,
            message=message,
            request_type="post",
        )
        self._status = response["status"]
        if self.status != TaskInstanceStatus.DONE:
            raise TransitionError(
                f"TaskInstance {self.task_instance_id} failed because it could not transition "
                f"to {TaskInstanceStatus.DONE} status. Current status is {self.status}."
            )

    def log_error(self, error_state: str, description: str) -> None:
        """Tell the JobStateManager that this task_instance has errored."""
        logger.info(f"Logging error for task_instance {self.task_instance_id}")

        message = {
            "error_state": error_state,
            "error_description": description,
            "stdout_log": self.command_stdout,
            "stderr_log": self.command_stderr,
            "stdout": self.stdout,
            "stderr": self.stderr,
            "nodename": self.nodename,
            "distributor_id": self.distributor_id,
        }

        app_route = f"/task_instance/{self.task_instance_id}/log_error_worker_node"
        _, response = self.requester.send_request(
            app_route=app_route,
            message=message,
            request_type="post",
        )
        self._status = response["status"]
        if self.status != error_state:
            raise TransitionError(
                f"TaskInstance {self.task_instance_id} failed because it could not transition "
                f"to {error_state} status. Current status is {self.status}."
            )

    def log_running(self) -> None:
        """Tell the JobStateManager that this task_instance is running.

        Update the report_by_date to be further in the future in case it gets reconciled
        immediately.
        """
        logger.info(f"Log running for task_instance {self.task_instance_id}")
        message = {
            "nodename": self.nodename,
            "process_group_id": str(self.process_group_id),
            "next_report_increment": (
                self._task_instance_heartbeat_interval
                * self._heartbeat_report_by_buffer
            ),
        }
        if self.distributor_id is not None:
            message["distributor_id"] = str(self.distributor_id)
        else:
            logger.info(
                "No distributor_id was found in the worker_node env at this time."
            )

        app_route = f"/task_instance/{self.task_instance_id}/log_running"
        _, response = self.requester.send_request(
            app_route=app_route,
            message=message,
            request_type="post",
        )

        kwargs = SerializeTaskInstance.kwargs_from_wire_worker_node(
            response["task_instance"]
        )
        self._status = kwargs.pop("status")
        self._command = kwargs.pop("command")
        task_name = kwargs.pop("name")
        self._stdout = self.cluster_interface.initialize_logfile(
            "stdout", kwargs.pop("stdout_dir"), task_name
        )
        self._stderr = self.cluster_interface.initialize_logfile(
            "stderr", kwargs.pop("stderr_dir"), task_name
        )
        self._command_add_env = {
            f"JOBMON_{k.upper()}": str(v) for k, v in kwargs.items()
        }
        self.last_heartbeat_time = time()
        if self.status != TaskInstanceStatus.RUNNING:
            raise TransitionError(
                f"TaskInstance {self.task_instance_id} failed because it could not transition "
                f"to {TaskInstanceStatus.RUNNING} status. Current status is {self.status}."
            )

    def log_report_by(self) -> None:
        """Log the heartbeat to show that the task instance is still alive."""
        logger.debug(f"Logging heartbeat for task_instance {self.task_instance_id}")
        message: Dict = {
            "next_report_increment": (
                self._task_instance_heartbeat_interval
                * self._heartbeat_report_by_buffer
            ),
            "stdout": self.stdout,
            "stderr": self.stderr,
        }
        if self.distributor_id is not None:
            message["distributor_id"] = str(self.distributor_id)
        else:
            logger.debug("No distributor_id was found in the sbatch env at this time")

        app_route = f"/task_instance/{self.task_instance_id}/log_report_by"
        _, response = self.requester.send_request(
            app_route=app_route,
            message=message,
            request_type="post",
        )

        self._status = response["status"]
        self.last_heartbeat_time = time()

        if self.status != TaskInstanceStatus.RUNNING:
            raise TransitionError(
                f"TaskInstance {self.task_instance_id} failed because it could not transition "
                f"to {TaskInstanceStatus.RUNNING} status. Current status is {self.status}."
            )

    def run(self) -> None:
        """This script executes on the target node and wraps the target application.

        Could be in any language, anything that can execute on linux. Similar to a stub or a
        container set ENV variables in case tasks need to access them.

        """
        # If it logs running and is not able to transition it raises TransitionError
        self.log_running()

        try:
            # run the command in a subprocess
            asyncio.run(self._run_cmd())

        # some other deployment unit transitioned task instance out of R state
        except RuntimeError as e:
            if isinstance(e.__cause__, TransitionError):
                msg = (
                    f"TaskInstance is in status '{self.status}'. Expected status 'R'."
                    f" Terminating command {self.command}."
                )
                logger.error(msg)

                # log an error with db if we are in K state
                if self.status == TaskInstanceStatus.KILL_SELF:
                    msg = (
                        f"Command: '{self.command}' got KILL_SELF event. Process shut down"
                        f" with exit code: '{self.command_returncode}'"
                    )
                    logger.error(msg)
                    self.log_error(TaskInstanceStatus.ERROR_FATAL, msg)

                # otherwise raise the error cause we are in trouble
                else:
                    raise e.__cause__
            else:
                raise

        # normal happy path
        else:
            if self.command_returncode == ReturnCodes.OK:
                logger.info(f"Command: {self.command}. Finished Successfully.")
                self.log_done()
            else:
                logger.info(
                    f"Command: {self.command} exited with signal {self.command_returncode}"
                )
                error_state, msg = self.cluster_interface.get_exit_info(
                    self.command_returncode, self.command_stderr
                )
                self.log_error(error_state, msg)

    def set_command_output(self, returncode: int, stdout: str, stderr: str) -> None:
        self._proc_returncode = returncode
        self._proc_stdout = stdout
        self._proc_stderr = stderr

    @staticmethod
    async def _communicate(
        async_stream: asyncio.StreamReader, output_stream: TextIO, chunk_size: int = 64
    ) -> str:
        mem_buffer = ""
        try:
            while True:
                # Read a chunk of data. If no data is returned, we've reached EOF.
                data_chunk = await async_stream.read(chunk_size)
                if not data_chunk:
                    break  # EOF reached

                # Attempt to decode and write the data chunk.
                try:
                    data_chunk_str = data_chunk.decode()
                    output_stream.write(data_chunk_str)
                    output_stream.flush()
                    mem_buffer += data_chunk_str
                    # Keep only the last 10k characters in memory.
                    mem_buffer = mem_buffer[-10000:]
                except UnicodeDecodeError:
                    pass  # Ignore decoding errors and continue reading the stream.
        except Exception as e:
            # Log unexpected errors. This could be any exception raised by
            # the reading or writing operations. Consider appending an error message
            # to `mem_buffer` to indicate that an error occurred.
            mem_buffer += "\n[Error reading stream: {}]".format(e)
        finally:
            # Ensure that the method always returns the buffer, even if an error occurred.
            return mem_buffer

    async def _process_poller(self, process: asyncio.subprocess.Process) -> int:
        keep_polling = True
        while keep_polling:
            time_till_next_heartbeat = self._task_instance_heartbeat_interval - (
                time() - self.last_heartbeat_time
            )
            try:
                await asyncio.wait_for(process.wait(), timeout=time_till_next_heartbeat)
                keep_polling = False
            except asyncio.TimeoutError:
                self.log_report_by()

        # keep typecheck happy
        returncode = process.returncode
        if returncode is None:
            raise AttributeError(
                "process finished polling but does not a a return code."
            )

        return returncode

    async def _run_cmd(self) -> None:
        # Copy the current environment variables and update them with additional settings
        env = os.environ.copy()
        env.update(self.command_add_env)

        # capture stdout and stderr for asynchronous reading
        process = await asyncio.create_subprocess_shell(
            self.command,
            env=env,
            stdout=asyncio.subprocess.PIPE,  # Captures stdout
            stderr=asyncio.subprocess.PIPE,  # Captures stderr
        )

        # Assert that stdout and stderr are not None for the type checker.
        assert process.stdout is not None
        assert process.stderr is not None

        # Initialize task variables to None. These will hold the asyncio tasks for
        # communicating with and monitoring the subprocess.
        stdout_task = stderr_task = heartbeat_task = None

        try:
            # Context manager to ensure file streams are properly closed after writing.
            with open(self.stdout, "w") as stdout_stream, open(
                self.stderr, "w"
            ) as stderr_stream:
                # Create asyncio tasks for reading subprocess stdout and stderr.
                stdout_task = asyncio.create_task(
                    self._communicate(process.stdout, stdout_stream)
                )
                stderr_task = asyncio.create_task(
                    self._communicate(process.stderr, stderr_stream)
                )
                # Task for monitoring the subprocess (e.g., for timeouts or heartbeats).
                heartbeat_task = asyncio.create_task(self._process_poller(process))

                # Await the completion of communication and monitoring tasks.
                # We gather all tasks together, ensuring they are completed before proceeding.
                valid_tasks = [stdout_task, stderr_task, heartbeat_task]
                await asyncio.gather(*valid_tasks)

        except Exception as e:
            # If an exception occurs, cancel all ongoing tasks to prevent dangling operations.
            tasks_to_cancel = [
                t for t in [stdout_task, stderr_task, heartbeat_task] if t is not None
            ]
            for task in tasks_to_cancel:
                task.cancel()
            # Await cancellation, ignoring any exceptions raised from cancellation.
            await asyncio.gather(*tasks_to_cancel, return_exceptions=True)

            # Attempt a graceful shutdown of the subprocess if it's still running.
            process.send_signal(signal.SIGINT)
            try:
                # Wait for the process to terminate, with a specified timeout.
                await asyncio.wait_for(
                    process.wait(), timeout=self._command_interrupt_timeout
                )
            except asyncio.TimeoutError:
                # Forcefully terminate the subprocess
                process.kill()
                await process.wait()

            raise RuntimeError(
                f"Failed to execute command '{self.command}': {e}"
            ) from e

        finally:
            # After tasks have been handled and the subprocess is ensured to be terminated,
            # retrieve the results from the tasks if they were successfully completed.
            stdout_result = stderr_result = ""
            if stdout_task and stdout_task.done():
                stdout_result = stdout_task.result()
            if stderr_task and stderr_task.done():
                stderr_result = stderr_task.result()

            # Ensure the subprocess is fully terminated before exiting this method.
            returncode = await process.wait()

            # Update the task instance with the final results of command execution.
            self.set_command_output(
                returncode=returncode,
                stdout=stdout_result,
                stderr=stderr_result,
            )



from __future__ import annotations

import asyncio
from collections import defaultdict
import itertools as it
import logging
import signal
import sys
import time
import traceback
from typing import (
    Any,
    Callable,
    Dict,
    Generator,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    Union,
)

import aiohttp

from jobmon.core import __version__
from jobmon.core.cluster_protocol import ClusterDistributor
from jobmon.core.configuration import JobmonConfig
from jobmon.core.constants import TaskInstanceStatus
from jobmon.core.exceptions import DistributorInterruptedError, InvalidResponse
from jobmon.core.requester import http_request_ok, Requester
from jobmon.core.serializers import SerializeTaskInstanceBatch
from jobmon.distributor.distributor_command import DistributorCommand
from jobmon.distributor.distributor_task_instance import DistributorTaskInstance
from jobmon.distributor.distributor_workflow_run import DistributorWorkflowRun
from jobmon.distributor.task_instance_batch import TaskInstanceBatch


logger = logging.getLogger(__name__)


class DistributorService:
    def __init__(
        self,
        cluster_interface: ClusterDistributor,
        requester: Optional[Requester] = None,
        workflow_run_heartbeat_interval: Optional[int] = None,
        task_instance_heartbeat_interval: Optional[int] = None,
        heartbeat_report_by_buffer: Optional[float] = None,
        distributor_poll_interval: Optional[int] = None,
        raise_on_error: bool = False,
    ) -> None:
        """Initialization of DistributorService."""
        # operational args
        config = JobmonConfig()
        if workflow_run_heartbeat_interval is None:
            self._workflow_run_heartbeat_interval = config.get_int(
                "heartbeat", "workflow_run_interval"
            )
        else:
            self._workflow_run_heartbeat_interval = workflow_run_heartbeat_interval
        if task_instance_heartbeat_interval is None:
            self._task_instance_heartbeat_interval = config.get_int(
                "heartbeat", "task_instance_interval"
            )
        else:
            self._task_instance_heartbeat_interval = task_instance_heartbeat_interval
        if heartbeat_report_by_buffer is None:
            self._heartbeat_report_by_buffer = config.get_float(
                "heartbeat", "report_by_buffer"
            )
        else:
            self._heartbeat_report_by_buffer = heartbeat_report_by_buffer
        if distributor_poll_interval is None:
            self._distributor_poll_interval = config.get_int(
                "distributor", "poll_interval"
            )
        else:
            self._distributor_poll_interval = distributor_poll_interval
        self.raise_on_error = raise_on_error

        # indexing of task instance by associated id
        self._task_instances: Dict[int, DistributorTaskInstance] = {}
        self._task_instance_batches: Dict[Tuple[int, int], TaskInstanceBatch] = {}

        # work queue
        self._distributor_commands: Iterator[DistributorCommand] = it.chain([])

        # indexing of task instances by status
        self._task_instance_status_map: Dict[str, Set[DistributorTaskInstance]] = {
            TaskInstanceStatus.QUEUED: set(),
            TaskInstanceStatus.INSTANTIATED: set(),
            TaskInstanceStatus.LAUNCHED: set(),
            TaskInstanceStatus.RUNNING: set(),
            TaskInstanceStatus.TRIAGING: set(),
            TaskInstanceStatus.KILL_SELF: set(),
            TaskInstanceStatus.NO_HEARTBEAT: set(),
        }
        # order through which we processes work
        gen_map: Dict[str, Callable[..., Generator[DistributorCommand, None, None]]] = {
            TaskInstanceStatus.QUEUED: self._check_queued_for_work,
            TaskInstanceStatus.INSTANTIATED: self._check_instantiated_for_work,
            TaskInstanceStatus.TRIAGING: self._check_triaging_for_work,
            TaskInstanceStatus.KILL_SELF: self._check_kill_self_for_work,
            TaskInstanceStatus.NO_HEARTBEAT: self._check_no_heartbeat_for_work,
        }
        self._command_generator_map = gen_map

        # syncronization timings
        self._last_heartbeat_time = time.time()

        # cluster API
        self.cluster_interface = cluster_interface

        # web service API
        if requester is None:
            self.requester = Requester.from_defaults()
        else:
            self.requester = requester

    @property
    def _next_report_increment(self) -> float:
        return self._heartbeat_report_by_buffer * self._task_instance_heartbeat_interval

    def set_workflow_run(self, workflow_run_id: int) -> None:
        workflow_run = DistributorWorkflowRun(workflow_run_id, self.requester)
        self.workflow_run = workflow_run
        self.workflow_run.transition_to_instantiated()

    def run(self) -> None:
        # start the cluster
        try:
            self._initialize_signal_handlers()
            self.cluster_interface.start()
            self.workflow_run.transition_to_launched()

            # signal via pipe that we are alive
            sys.stderr.write("ALIVE")
            sys.stderr.flush()

            done: List[str] = []
            todo = [
                TaskInstanceStatus.QUEUED,
                TaskInstanceStatus.INSTANTIATED,
                TaskInstanceStatus.LAUNCHED,
                TaskInstanceStatus.RUNNING,
                TaskInstanceStatus.TRIAGING,
                TaskInstanceStatus.KILL_SELF,
                TaskInstanceStatus.NO_HEARTBEAT,
            ]
            while True:
                # loop through all statuses and do as much work as we can till the heartbeat
                time_till_next_heartbeat = self._workflow_run_heartbeat_interval - (
                    time.time() - self._last_heartbeat_time
                )

                while todo and time_till_next_heartbeat > 0:
                    # log when this status started
                    start_time = time.time()

                    # remove status from todo and add to done
                    status = todo.pop(0)

                    # refresh internal state from db
                    self.refresh_status_from_db(status)

                    # how long the heartbeat took
                    refresh_time = time.time()
                    time_till_next_heartbeat -= refresh_time - start_time

                    if status in self._command_generator_map.keys():
                        # process any work
                        self.process_status(status, time_till_next_heartbeat)
                        # how long the full status took
                        end_time = time.time()
                        time_till_next_heartbeat -= end_time - refresh_time

                    else:
                        end_time = refresh_time

                    done.append(status)
                    logger.info(
                        f"Status processing for status={status} took "
                        f"{int((end_time - start_time))}s."
                    )

                # append done work to the end of the work order
                todo += done
                done = []

                if time_till_next_heartbeat > 0:
                    time.sleep(time_till_next_heartbeat)

                self.log_task_instance_report_by_date()

        except DistributorInterruptedError:
            logger.info("Interrupt received!")
        except Exception as e:
            logger.exception(e)
            raise
        finally:
            # stop distributor
            self.cluster_interface.stop()

            # signal via pipe that we are shutdown
            sys.stderr.write("SHUTDOWN")
            sys.stderr.flush()

    def process_status(self, status: str, timeout: Union[int, float] = -1) -> None:
        """Processes commands until all work is done or timeout is reached.

        Args:
            status: which status to process work for.
            timeout: time until we stop processing. -1 means process till no more work
        """
        start = time.time()

        # generate new distributor commands from this status
        command_generator_callable = self._command_generator_map[status]
        command_generator = command_generator_callable()
        self._distributor_commands = it.chain(command_generator)

        # this way we always process at least 1 command
        keep_iterating = True
        while keep_iterating:
            # run commands
            try:
                # get next command
                distributor_command = next(self._distributor_commands)
                distributor_command(self.raise_on_error)

                # if we need a status sync close the main generator. we will process remaining
                # transactions, but nothing new from the generator
                if not ((time.time() - start) < timeout or timeout == -1):
                    command_generator.close()

            except StopIteration:
                # stop processing commands if we are out of commands
                keep_iterating = False

        # update the state map
        task_instances = self._task_instance_status_map.pop(status)
        self._task_instance_status_map[status] = set()
        for task_instance in task_instances:
            self._task_instance_status_map[task_instance.status].add(task_instance)

    def instantiate_task_instances(
        self, task_instances: List[DistributorTaskInstance]
    ) -> None:
        app_route = "/task_instance/instantiate_task_instances"
        _, result = self.requester.send_request(
            app_route=app_route,
            message={
                "task_instance_ids": [
                    task_instance.task_instance_id for task_instance in task_instances
                ]
            },
            request_type="post",
        )

        # construct batch. associations are made inside batch init
        for batch in result["task_instance_batches"]:
            task_instance_batch_kwargs = SerializeTaskInstanceBatch.kwargs_from_wire(
                batch
            )

            array_id = task_instance_batch_kwargs["array_id"]
            batch_number = task_instance_batch_kwargs["array_batch_num"]
            try:
                task_instance_batch = self._task_instance_batches[
                    (array_id, batch_number)
                ]
            except KeyError:
                task_instance_batch = TaskInstanceBatch(
                    array_id=array_id,
                    array_name=task_instance_batch_kwargs["array_name"],
                    array_batch_num=batch_number,
                    task_resources_id=task_instance_batch_kwargs["task_resources_id"],
                    requester=self.requester,
                )
                self._task_instance_batches[(array_id, batch_number)] = (
                    task_instance_batch
                )

            for task_instance_id in task_instance_batch_kwargs["task_instance_ids"]:
                task_instance = self._task_instances[task_instance_id]
                task_instance.status = TaskInstanceStatus.INSTANTIATED
                task_instance_batch.add_task_instance(task_instance)

    def launch_task_instance_batch(
        self, task_instance_batch: TaskInstanceBatch
    ) -> None:
        self._task_instance_batches.pop(
            (task_instance_batch.array_id, task_instance_batch.batch_number)
        )
        # record batch info in db
        task_instance_batch.prepare_task_instance_batch_for_launch()

        # build worker node command
        command = self.cluster_interface.build_worker_node_command(
            task_instance_id=None,
            array_id=task_instance_batch.array_id,
            batch_number=task_instance_batch.batch_number,
        )
        distributor_commands: List[DistributorCommand] = []

        try:
            # submit array to distributor
            distributor_id_map = (
                self.cluster_interface.submit_array_to_batch_distributor(
                    command=command,
                    name=task_instance_batch.submission_name,
                    requested_resources=task_instance_batch.requested_resources,
                    array_length=len(task_instance_batch.task_instances),
                )
            )
            task_instance_batch.set_distributor_ids(distributor_id_map)

        except NotImplementedError:
            # create DistributorCommands to submit the launch if array isn't implemented
            for task_instance in task_instance_batch.task_instances:
                distributor_command = DistributorCommand(
                    self.launch_task_instance,
                    task_instance,
                )
                distributor_commands.append(distributor_command)

        except Exception as e:
            # if other error, transition to No ID status
            stack_trace = traceback.format_exc()
            logger.exception(e)
            for task_instance in task_instance_batch.task_instances:
                distributor_command = DistributorCommand(
                    task_instance.transition_to_no_distributor_id,
                    no_id_err_msg=stack_trace,
                )
                distributor_commands.append(distributor_command)

        else:
            # if successful log a transition to launched
            launch_command = DistributorCommand(
                task_instance_batch.transition_to_launched, self._next_report_increment
            )
            # Log the distributor IDs
            log_distributor_ids_command = DistributorCommand(
                task_instance_batch.log_distributor_ids
            )

            distributor_commands.append(launch_command)
            distributor_commands.append(log_distributor_ids_command)

        finally:
            self._distributor_commands = it.chain(
                distributor_commands, self._distributor_commands
            )

    def launch_task_instance(self, task_instance: DistributorTaskInstance) -> None:
        """Submits a task instance on a given distributor.

        Adds the new task instance to self.submitted_or_running_task_instances.
        """
        # load resources
        try:
            requested_resources = task_instance.batch.requested_resources
        except AttributeError:
            task_instance.batch.load_requested_resources()
            requested_resources = task_instance.batch.requested_resources

        # Fetch the worker node command
        command = self.cluster_interface.build_worker_node_command(
            task_instance_id=task_instance.task_instance_id
        )

        # Submit to batch distributor
        try:
            distributor_id = self.cluster_interface.submit_to_batch_distributor(
                command=command,
                name=task_instance.submission_name,
                requested_resources=requested_resources,
            )
        except Exception as e:
            stack_trace = traceback.format_exc()
            logger.exception(e)
            task_instance.transition_to_no_distributor_id(no_id_err_msg=stack_trace)

        else:
            # move from register queue to launch queue
            task_instance.transition_to_launched(
                distributor_id, self._next_report_increment
            )

    def triage_error(self, task_instance: DistributorTaskInstance) -> None:
        """Triage a running task instance that has missed a heartbeat.

        Allowed transitions are (R, U, Z, F)
        """
        r_value, r_msg = self.cluster_interface.get_remote_exit_info(
            task_instance.distributor_id
        )
        task_instance.transition_to_error(r_msg, r_value)

    def kill_self_batch(self, task_instance_batch: TaskInstanceBatch) -> None:
        """Terminate all TIs in this batch.

        Args:
            task_instance_batch: The batch of task instances to terminate.
        """
        # 1) Collect the distributor IDs to terminate
        distributor_ids = [
            ti.distributor_id
            for ti in task_instance_batch.task_instances
            if ti.distributor_id is not None
        ]

        # 2) If there are jobs to terminate, call the cluster
        if distributor_ids:
            self.cluster_interface.terminate_task_instances(distributor_ids)

        # 3) Mark them as killed in the DB
        task_instance_batch.transition_to_killed()

    def no_heartbeat_error(self, task_instance: DistributorTaskInstance) -> None:
        """Move a task instance in NO_HEARTBEAT state to a recoverable error state.

        This signal is sent from the swarm in the event a task instance in LAUNCHED state
        fails to log a heartbeat, either due to the distributor failing to log a heartbeat
        batch or due to the worker node failing to start up properly.

        ERROR state allows for a retry, so that a new task instance can attempt to run.
        """
        task_instance.transition_to_error(
            "Task instance never reported a heartbeat after scheduling. Will retry",
            TaskInstanceStatus.ERROR,
        )

    def log_task_instance_report_by_date(self) -> None:
        """Log the heartbeat to show that the task instance is still alive."""
        task_instances_launched = self._task_instance_status_map[
            TaskInstanceStatus.LAUNCHED
        ]
        submitted_or_running = self.cluster_interface.get_submitted_or_running(
            [x.distributor_id for x in task_instances_launched]
        )

        task_instance_ids_to_heartbeat: List[int] = []
        for task_instance_launched in task_instances_launched:
            if task_instance_launched.distributor_id in submitted_or_running:
                task_instance_ids_to_heartbeat.append(
                    task_instance_launched.task_instance_id
                )

        if any(task_instance_ids_to_heartbeat):
            # Create batches of task instance IDs
            chunk_size = 500
            task_instance_batches = [
                task_instance_ids_to_heartbeat[i : i + chunk_size]
                for i in range(0, len(task_instance_ids_to_heartbeat), chunk_size)
            ]

            # Send heartbeat for each batch
            asyncio.run(self._log_heartbeats(task_instance_batches))

        self._last_heartbeat_time = time.time()

    async def _log_heartbeats(self, task_instance_batches: List[List[int]]) -> None:
        """Create a task for each batch of task instances to send heartbeat."""
        async with aiohttp.ClientSession(self.requester.base_url) as session:
            heartbeat_tasks = [
                asyncio.create_task(self._log_heartbeat_by_batch(session, batch))
                for batch in task_instance_batches
            ]
            await asyncio.gather(*heartbeat_tasks)

    async def _log_heartbeat_by_batch(
        self, session: aiohttp.ClientSession, task_instance_ids_to_heartbeat: List[int]
    ) -> None:
        """Send heartbeat for a batch of task instances."""
        message: Dict = {
            "next_report_increment": self._next_report_increment,
            "task_instance_ids": task_instance_ids_to_heartbeat,
        }
        app_route = f"{self.requester.route_prefix}/task_instance/log_report_by/batch"

        # Super basic retrying logic, to avoid fussing with tenacity logic.
        # TODO: Factor out into an asynchronous requester

        max_attempts, wait_time = 10, 1.5

        while max_attempts > 0:
            async with session.post(
                app_route,
                json=message,
                params={"client_jobmon_version": __version__},
                headers={"Content-Type": "application/json"},
            ) as response:
                return_code = response.status
                response_text = await response.text()

            if 499 < return_code < 600:
                logger.warning(
                    f"Got HTTP status_code={return_code} from server. "
                    f"app_route: {app_route}."
                )
            elif return_code == 423:
                logger.info(
                    f"Got HTTP status_code=423 from server. app_route: {app_route}. "
                    f"Retrying as per design."
                )
            else:
                break

            max_attempts -= 1
            await asyncio.sleep(wait_time)
            wait_time *= wait_time

        if not http_request_ok(return_code):
            raise InvalidResponse(
                f"Unexpected status code {return_code} from POST "
                f"request through route {app_route}. Expected "
                f"code 200. Response content: {response_text}"
            )

    def _initialize_signal_handlers(self) -> None:
        def handle_sighup(signal: int, frame: Any) -> None:
            raise DistributorInterruptedError("Got signal SIGHUP.")

        def handle_sigterm(signal: int, frame: Any) -> None:
            raise DistributorInterruptedError("Got signal SIGTERM.")

        def handle_sigint(signal: int, frame: Any) -> None:
            pass

        signal.signal(signal.SIGTERM, handle_sigterm)
        signal.signal(signal.SIGHUP, handle_sighup)
        signal.signal(signal.SIGINT, handle_sigint)

    def refresh_status_from_db(self, status: str) -> None:
        """Got to DB to check the list tis status."""
        message = {
            "task_instance_ids": [
                task_instance.task_instance_id
                for task_instance in self._task_instance_status_map[status]
            ],
            "status": status,
        }
        app_route = f"/workflow_run/{self.workflow_run.workflow_run_id}/sync_status"
        _, result = self.requester.send_request(
            app_route=app_route, message=message, request_type="post"
        )

        # mutate the statuses and update the status map
        status_updates: Dict[str, List[int]] = result["status_updates"]
        for new_status, task_instance_ids in status_updates.items():
            for task_instance_id in task_instance_ids:
                try:
                    task_instance = self._task_instances[task_instance_id]

                except KeyError:
                    task_instance = DistributorTaskInstance(
                        task_instance_id,
                        self.workflow_run.workflow_run_id,
                        new_status,
                        self.requester,
                    )
                    self._task_instance_status_map[task_instance.status].add(
                        task_instance
                    )
                    self._task_instances[task_instance.task_instance_id] = task_instance

                else:
                    # remove from old status set
                    previous_status = task_instance.status
                    self._task_instance_status_map[previous_status].remove(
                        task_instance
                    )

                    try:
                        self._task_instance_status_map[new_status].add(task_instance)
                        # change to new status and move to new set
                        task_instance.status = new_status

                    except KeyError:
                        # If the task instance is in a terminal state, e.g. D, E, etc.,
                        # expire it from the distributor
                        del self._task_instances[task_instance_id]

    def _check_queued_for_work(self) -> Generator[DistributorCommand, None, None]:
        queued_task_instances = list(
            self._task_instance_status_map[TaskInstanceStatus.QUEUED]
        )
        queued_task_instances.sort()
        chunk_size = 500
        while queued_task_instances:
            ti_list = queued_task_instances[:chunk_size]
            queued_task_instances = queued_task_instances[chunk_size:]
            yield DistributorCommand(self.instantiate_task_instances, ti_list)

    def _check_instantiated_for_work(self) -> Generator[DistributorCommand, None, None]:
        # compute the task_instances that can be launched
        instantiated_task_instances = list(
            self._task_instance_status_map[TaskInstanceStatus.INSTANTIATED]
        )
        task_instance_batches = set(
            [task_instance.batch for task_instance in instantiated_task_instances]
        )

        for batch in task_instance_batches:
            yield DistributorCommand(self.launch_task_instance_batch, batch)

    def _check_triaging_for_work(self) -> Generator[DistributorCommand, None, None]:
        """Handle TIs in TRIAGING state.

        For TaskInstances with TRIAGING status, check the nature of no heartbeat,
        and change the statuses accordingly.
        """
        triaging_task_instances = self._task_instance_status_map[
            TaskInstanceStatus.TRIAGING
        ]
        for task_instance in triaging_task_instances:
            yield DistributorCommand(self.triage_error, task_instance)

    def _check_kill_self_for_work(self) -> Generator[DistributorCommand, None, None]:
        """Handle TIs in KILL_SELF state, grouped by their TaskInstanceBatch."""
        kill_self_task_instances = list(
            self._task_instance_status_map[TaskInstanceStatus.KILL_SELF]
        )

        # Group TIs by their batch
        batch_map = defaultdict(list)
        for ti in kill_self_task_instances:
            batch_map[ti.batch].append(ti)

        for batch_obj, _ in batch_map.items():
            # If you'd like to verify they still have KILL_SELF status, etc., do it here.
            yield DistributorCommand(self.kill_self_batch, batch_obj)

    def _check_no_heartbeat_for_work(self) -> Generator[DistributorCommand, None, None]:
        """Handle TIs in NO_HEARTBEAT state.

        For TaskInstances with NO_HEARTBEAT status, move to an error recoverable state
        """
        no_heartbeat_task_instances = self._task_instance_status_map[
            TaskInstanceStatus.NO_HEARTBEAT
        ]

        for task_instance in no_heartbeat_task_instances:
            yield DistributorCommand(self.no_heartbeat_error, task_instance)



"""Command line interface for Execution."""

import argparse
import importlib
import importlib.machinery
import importlib.util
import logging
import os
import sys
from typing import Optional

from jobmon.core.cli import CLI
from jobmon.core.task_generator import TaskGenerator

logger = logging.getLogger(__name__)


class WorkerNodeCLI(CLI):
    """Command line interface for WorkderNode."""

    def __init__(self) -> None:
        """Initialization of the worker node CLI."""
        self.parser = argparse.ArgumentParser("jobmon worker_node CLI")
        self._subparsers = self.parser.add_subparsers(
            dest="sub_command", parser_class=argparse.ArgumentParser
        )

        self._add_worker_node_job_parser()
        self._add_worker_node_array_parser()
        self._add_run_task_generator_parser()

    def run_task_instance_job(self, args: argparse.Namespace) -> int:
        """Configuration for the jobmon worker node."""
        from jobmon.core.exceptions import ReturnCodes
        from jobmon.worker_node import __version__
        from jobmon.worker_node.worker_node_factory import WorkerNodeFactory

        if __version__ != args.expected_jobmon_version:
            msg = (
                f"Your expected Jobmon version is {args.expected_jobmon_version} and your "
                f"worker node is using {__version__}. Please check your bash profile "
            )
            logger.error(msg)
            sys.exit(ReturnCodes.WORKER_NODE_ENV_FAILURE)

        worker_node_factory = WorkerNodeFactory(cluster_name=args.cluster_name)
        worker_node_task_instance = worker_node_factory.get_job_task_instance(
            task_instance_id=args.task_instance_id
        )
        worker_node_task_instance.configure_logging()
        try:
            worker_node_task_instance.run()
        except Exception as e:
            logger.error(e)
            sys.exit(ReturnCodes.WORKER_NODE_CLI_FAILURE)

        return worker_node_task_instance.command_returncode

    def run_task_instance_array(self, args: argparse.Namespace) -> int:
        """Configuration for the jobmon worker node."""
        from jobmon.core.exceptions import ReturnCodes
        from jobmon.worker_node import __version__
        from jobmon.worker_node.worker_node_factory import WorkerNodeFactory

        if __version__ != args.expected_jobmon_version:
            msg = (
                f"Your expected Jobmon version is {args.expected_jobmon_version} and your "
                f"worker node is using {__version__}. Please check your bash profile "
            )
            logger.error(msg)
            sys.exit(ReturnCodes.WORKER_NODE_ENV_FAILURE)

        worker_node_factory = WorkerNodeFactory(cluster_name=args.cluster_name)
        worker_node_task_instance = worker_node_factory.get_array_task_instance(
            array_id=args.array_id,
            batch_number=args.batch_number,
        )
        worker_node_task_instance.configure_logging()

        try:
            worker_node_task_instance.run()
        except Exception as e:
            logger.error(e)
            sys.exit(ReturnCodes.WORKER_NODE_CLI_FAILURE)

        return worker_node_task_instance.command_returncode

    def run_task_generator(self, args: argparse.Namespace) -> int:
        from jobmon.core.exceptions import ReturnCodes

        # Import the module and get the task generator we've been pointed to, raise an error
        # if it's not a TaskGenerator
        # if the user used the --module_dir flag, add the module directory to the path
        if args.module_source_path:
            # Create a module spec from the source file
            loader = importlib.machinery.SourceFileLoader(
                args.module_name, os.path.expanduser(args.module_source_path)
            )
            spec = importlib.util.spec_from_loader(loader.name, loader)
            # Create a new module based on the spec
            mod = importlib.util.module_from_spec(spec)  # type: ignore
            # Add the module to sys.modules
            sys.modules[args.module_name] = mod
            loader.exec_module(mod)
        else:
            mod = importlib.import_module(args.module_name)
        task_generator = getattr(mod, args.func_name)
        if not isinstance(task_generator, TaskGenerator):
            raise ValueError(
                f"{args.module_name}:{args.func_name} doesn't point to a runnable jobmon task."
            )

        # if the user used the --arghelp flag, print the help message for the task generator
        if args.arghelp:
            print(task_generator.help())
            return ReturnCodes.OK
        try:
            task_generator.run(args.args)
            return ReturnCodes.OK
        except Exception as e:
            print(e)
            raise e

    def _add_run_task_generator_parser(self) -> None:
        generator_parser = self._subparsers.add_parser("task_generator")
        generator_parser.set_defaults(func=self.run_task_generator)
        generator_parser.add_argument(
            "--module_name",
            help="name of the module containing the TaskGenerator",
            required=True,
        )
        generator_parser.add_argument(
            "--func_name",
            type=str,
            help="the name of the function which has been turned into a TaskGenerator",
            required=True,
        )
        generator_parser.add_argument(
            "--args",
            type=str,
            help="Followed by the key=value; .\n"
            "For example: \n"
            "If you method has two argument: def func(foo: int, bar: str), pass\n"
            "    --args foo=1 --args bar='test'\n"
            "If you method has two argument: def func(foo: int, bar: List[str), pass\n"
            "    --args foo=1 --args bar=[a,b]\n",
            required=False,
            action="append",
        )
        generator_parser.add_argument(
            "--arghelp",
            type=str,
            help="Show the help message for the task generator. For example: --arghelp",
            required=False,
        )
        generator_parser.add_argument(
            "--module_source_path",
            type=str,
            help="The directory the module source code located; "
            "you do not need this if the module is installed in your system.",
            required=False,
        )

    def _add_worker_node_job_parser(self) -> None:
        job_parser = self._subparsers.add_parser("worker_node_job")
        job_parser.set_defaults(func=self.run_task_instance_job)
        job_parser.add_argument(
            "--task_instance_id",
            help="task_instance_id of the work node.",
            required=False,
        )
        job_parser.add_argument(
            "--cluster_name",
            type=str,
            help="cluster_name of the work node.",
            required=True,
        )
        job_parser.add_argument(
            "--expected_jobmon_version",
            type=str,
            help="expected_jobmon_version of the work node.",
            required=True,
        )

    def _add_worker_node_array_parser(self) -> None:
        array_parser = self._subparsers.add_parser("worker_node_array")
        array_parser.set_defaults(func=self.run_task_instance_array)
        array_parser.add_argument(
            "--array_id",
            help="array_id of the worker node if this is an array task.",
            required=False,
        )
        array_parser.add_argument(
            "--batch_number",
            help="batch number of the array this task instance is associated with.",
            required=False,
        )
        array_parser.add_argument(
            "--cluster_name",
            type=str,
            help="cluster_name of the work node.",
            required=True,
        )
        array_parser.add_argument(
            "--expected_jobmon_version",
            type=str,
            help="expected_jobmon_version of the work node.",
            required=True,
        )


def run(argstr: Optional[str] = None) -> None:
    """Entrypoint to create WorkerNode CLI."""
    cli = WorkerNodeCLI()
    cli.main(argstr)



"""Inteface definition for jobmon executor plugins."""

from __future__ import annotations

import importlib
import pkgutil
from typing import Any, Dict, Type

from jobmon.core.cluster_protocol import (
    ClusterDistributor,
    ClusterQueue,
    ClusterWorkerNode,
)
import jobmon.plugins


class ClusterType:
    _interface = [
        "get_cluster_queue_class",
        "get_cluster_distributor_class",
        "get_cluster_worker_node_class",
    ]

    _cache: Dict[str, ClusterType] = {}

    def __new__(cls, *args: str, **kwds: str) -> ClusterType:
        key = args[0] if args else kwds["cluster_type_name"]
        inst = cls._cache.get(key, None)
        if inst is None:
            inst = super(ClusterType, cls).__new__(cls)
            inst.__init__(key)  # type: ignore
            cls._cache[key] = inst
        return inst

    def __init__(self, cluster_type_name: str) -> None:
        """Initialization of ClusterType object."""
        self.cluster_type_name = cluster_type_name

        self._plugins = {
            name: importlib.import_module(name)
            for finder, name, ispkg in pkgutil.iter_modules(
                jobmon.plugins.__path__, jobmon.plugins.__name__ + "."
            )
        }
        self._package_location = ""

    @property
    def plugin(self) -> Any:
        """If the cluster is bound, return the cluster interface for the type of cluster."""
        module = self._plugins.get(f"jobmon.plugins.{self.cluster_type_name}")
        if module is not None:
            msg = ""
            for func in self._interface:
                if not hasattr(module, func):
                    msg += f"Required function {func} missing from plugin interface. \n"
            if msg:
                raise AttributeError(
                    f"Invalid jobmon plugin {self.cluster_type_name}" + msg
                )
        else:
            msg = f"Interface not found for cluster_type_name={self.cluster_type_name}"
            raise ValueError(msg)
        return module

    @property
    def cluster_queue_class(self) -> Type[ClusterQueue]:
        return self.plugin.get_cluster_queue_class()

    @property
    def cluster_distributor_class(self) -> Type[ClusterDistributor]:
        return self.plugin.get_cluster_distributor_class()

    @property
    def cluster_worker_node_class(self) -> Type[ClusterWorkerNode]:
        return self.plugin.get_cluster_worker_node_class()



"""Dummy distributor that runs one task at a time."""

from collections import OrderedDict
import logging
import os
import random
import shutil
from typing import Any, Dict, List, Optional, Set, Tuple, Union

from jobmon.core.cluster_protocol import ClusterDistributor, ClusterWorkerNode
from jobmon.core.constants import TaskInstanceStatus
from jobmon.worker_node.cli import WorkerNodeCLI
from jobmon.worker_node.worker_node_factory import WorkerNodeFactory


logger = logging.getLogger(__name__)


class LimitedSizeDict(OrderedDict):
    """Dictionary for exit info."""

    def __init__(self, *args: int, **kwds: int) -> None:
        """Initialization of LimitedSizeDict."""
        self.size_limit = kwds.pop("size_limit", None)
        OrderedDict.__init__(self, *args, **kwds)
        self._check_size_limit()

    def __setitem__(self, key: Any, value: Any) -> None:
        """Set item in dict."""
        OrderedDict.__setitem__(self, key, value)
        self._check_size_limit()

    def _check_size_limit(self) -> None:
        if self.size_limit is not None:
            while len(self) > self.size_limit:
                self.popitem(last=False)


class DummyDistributor(ClusterDistributor):
    """Executor to run tasks one at a time."""

    def __init__(
        self,
        cluster_name: str,
        exit_info_queue_size: int = 1000,
        *args: tuple,
        **kwargs: dict,
    ) -> None:
        """Initialization of the dummy distributor.

        Args:
            cluster_name (str): name of the cluster
            exit_info_queue_size (int): how many exit codes to retain
        """
        self.started = False

        self._cluster_name = cluster_name
        worker_node_entry_point = shutil.which("worker_node_entry_point")
        if not worker_node_entry_point:
            raise ValueError("worker_node_entry_point can't be found.")
        self._worker_node_entry_point = worker_node_entry_point

        self._next_distributor_id = 1
        self._exit_info = LimitedSizeDict(size_limit=exit_info_queue_size)

    @property
    def worker_node_entry_point(self) -> str:
        """Path to jobmon worker_node_entry_point."""
        return self._worker_node_entry_point

    @property
    def cluster_name(self) -> str:
        """Return the name of the cluster type."""
        return self._cluster_name

    def start(self) -> None:
        """Start the distributor."""
        self.started = True

    def stop(self) -> None:
        """Stop the distributor."""
        self.started = False

    def get_queueing_errors(self, distributor_ids: List[str]) -> Dict[str, str]:
        """Get the task instances that have errored out."""
        raise NotImplementedError

    def get_array_queueing_errors(
        self, distributor_id: Union[int, str]
    ) -> Dict[Union[int, str], str]:
        raise NotImplementedError

    def get_remote_exit_info(self, distributor_id: str) -> Tuple[str, str]:
        """Get exit info from task instances that have run."""
        return TaskInstanceStatus.UNKNOWN_ERROR, "Whatever"

    def get_submitted_or_running(
        self, distributor_ids: Optional[List[str]] = None
    ) -> Set[str]:
        """Check status of running task."""
        running = os.environ.get("JOB_ID", "")
        return {running}

    def terminate_task_instances(self, distributor_ids: List[str]) -> None:
        """Terminate task instances.

        If implemented, return a list of (task_instance_id, hostname) tuples for any
        task_instances that are terminated.
        """
        logger.warning(
            "terminate_task_instances not implemented by ClusterDistributor: "
            f"{self.__class__.__name__}"
        )

    def submit_to_batch_distributor(
        self,
        command: str,
        name: str,
        requested_resources: Dict[str, Any],
    ) -> str:
        """Run a fake execution of the task.

        In a real executor, this is where submission to the cluster would happen.
        Here, since it's a dummy executor, we just get a random number and empty
        file paths.
        """
        logger.debug("This is the Dummy Distributor")
        # even number for non array tasks
        distributor_id = random.randint(1, int(1e6)) * 2
        os.environ["JOB_ID"] = str(distributor_id)

        cli = WorkerNodeCLI()
        args = cli.parse_args(command)

        worker_node_factory = WorkerNodeFactory(cluster_name=args.cluster_name)
        # Do not do ANY logging at all
        worker_node_task_instance = worker_node_factory.get_job_task_instance(
            task_instance_id=args.task_instance_id
        )
        # Log running, log done, and exit
        worker_node_task_instance.log_running()
        worker_node_task_instance.set_command_output(0, "", "")
        worker_node_task_instance.log_done()

        return str(distributor_id)


class DummyWorkerNode(ClusterWorkerNode):
    """Get Executor Info for a Task Instance."""

    def __init__(self) -> None:
        """Initialization of the dummy executor worker node."""
        self._distributor_id: Optional[str] = None
        self._logfile_template = {
            "stdout": "{root}/{name}.o{job_id}",
            "stderr": "{root}/{name}.e{job_id}",
        }

    @property
    def distributor_id(self) -> Optional[str]:
        """Distributor id of the task."""
        """Distributor id of the task."""
        if self._distributor_id is None:
            jid = os.environ.get("JOB_ID")
            if jid:
                self._distributor_id = jid
        return self._distributor_id

    def initialize_logfile(self, log_type: str, log_dir: str, name: str) -> str:
        if log_dir:
            logpath = self._logfile_template[log_type].format(
                root=log_dir, name=name, job_id=self.distributor_id
            )
        else:
            logpath = "/dev/null"
        return logpath

    @staticmethod
    def get_exit_info(exit_code: int, error_msg: str) -> Tuple[str, str]:
        """Exit info, error message."""
        return TaskInstanceStatus.ERROR, error_msg

    @staticmethod
    def get_usage_stats() -> Dict:
        """Usage information specific to the exector."""
        return {}



"""Cluster objects define where a user wants their tasks run. e.g. UGE, Azure, Seq."""

from __future__ import annotations

import logging
from typing import Any, Dict, Optional

from jobmon.core.cluster_protocol import (
    ClusterDistributor,
    ClusterQueue,
    ClusterWorkerNode,
)
from jobmon.core.cluster_type import ClusterType
from jobmon.core.requester import Requester
from jobmon.core.serializers import SerializeCluster, SerializeQueue


logger = logging.getLogger(__name__)


class Cluster:
    """Cluster objects define where a user wants their tasks run. e.g. UGE, Azure, Seq."""

    def __init__(
        self, cluster_name: str, requester: Optional[Requester] = None
    ) -> None:
        """Initialization of Cluster."""
        self.cluster_name = cluster_name

        if requester is None:
            requester = Requester.from_defaults()
        self.requester = requester

        self.queues: Dict[str, ClusterQueue] = {}

    @classmethod
    def get_cluster(
        cls: Any, cluster_name: str, requester: Optional[Requester] = None
    ) -> Cluster:
        """Get a bound instance of a Cluster.

        Args:
            cluster_name: the name of the cluster
            requester (Requester): requester object to connect to FastApi service.
        """
        cluster = cls(cluster_name, requester)
        cluster.bind()
        return cluster

    def bind(self) -> None:
        """Bind Cluster to the database, getting an id back."""
        app_route = f"/cluster/{self.cluster_name}"
        _, response = self.requester.send_request(
            app_route=app_route, message={}, request_type="get"
        )
        cluster_kwargs = SerializeCluster.kwargs_from_wire(response["cluster"])

        self._cluster_id = cluster_kwargs["id"]
        cluster_type = ClusterType(cluster_kwargs["cluster_type_name"])
        self._cluster_type = cluster_type
        self._connection_parameters = cluster_kwargs["connection_parameters"]

    @property
    def connection_parameters(self) -> Dict:
        """The connection parameters."""
        return self._connection_parameters

    @property
    def is_bound(self) -> bool:
        """If the Cluster has been bound to the database."""
        return hasattr(self, "_cluster_id")

    @property
    def id(self) -> int:
        """Unique id from database if Cluster has been bound."""
        if not self.is_bound:
            raise AttributeError("Cannot access id until Cluster is bound to database")
        return self._cluster_id

    def get_worker_node(self) -> ClusterWorkerNode:
        """Get the cluster specific worker_node interface."""
        cluster_worker_node_class = self._cluster_type.cluster_worker_node_class
        return cluster_worker_node_class()

    def get_distributor(self) -> ClusterDistributor:
        """Get the cluster specific distributor interface."""
        # TODO: read in cluster args from config here?
        distributor_class = self._cluster_type.cluster_distributor_class
        return distributor_class(self.cluster_name, **self._connection_parameters)

    def get_queue(self, queue_name: str) -> ClusterQueue:
        """Get the ClusterQueue object associated with a given queue_name.

        Checks if queue object is in the cache, if it's not it will query the database and add
        the queue object to the cache.

        Args:
            queue_name: name of the queue you want.
        """
        # this is cached so should be fast
        try:
            queue = self.queues[queue_name]
        except KeyError:
            queue_class = self._cluster_type.cluster_queue_class
            app_route = f"/cluster/{self.id}/queue/{queue_name}"
            _, response = self.requester.send_request(
                app_route=app_route, message={}, request_type="get"
            )
            queue_kwargs = SerializeQueue.kwargs_from_wire(response["queue"])
            queue = queue_class(**queue_kwargs)
            self.queues[queue_name] = queue

        return queue



"""jobmon built-in."""

from importlib.metadata import version
from typing import Type

from jobmon.core.cluster_protocol import (
    ClusterDistributor,
    ClusterQueue,
    ClusterWorkerNode,
)

__version__ = version("jobmon_core")


def get_cluster_queue_class() -> Type[ClusterQueue]:
    """Return the queue class for the Multiprocess executor."""
    from jobmon.plugins.multiprocess.multiproc_queue import MultiprocessQueue

    return MultiprocessQueue


def get_cluster_distributor_class() -> Type[ClusterDistributor]:
    """Return the cluster distributor for the Multiprocess executor."""
    from jobmon.plugins.multiprocess.multiproc_distributor import (
        MultiprocessDistributor,
    )

    return MultiprocessDistributor


def get_cluster_worker_node_class() -> Type[ClusterWorkerNode]:
    """Return the cluster worker node class for the Multiprocess executor."""
    from jobmon.plugins.multiprocess.multiproc_distributor import (
        MultiprocessWorkerNode,
    )

    return MultiprocessWorkerNode



"""Parse configuration options and set them to be used throughout the Jobmon Architecture."""

import argparse
import os
from pathlib import Path
from typing import Any, Dict, Optional, Union

from dotenv import load_dotenv
import yaml

from jobmon.core import CONFIG_FILE_FROM_INSTALLER_PLUGIN
from jobmon.core.cli import CLI
from jobmon.core.exceptions import ConfigError

load_dotenv()

DEFAULTS_FILE_NAME = "defaults.yaml"
DEFAULTS_FILE = Path(__file__).parent / DEFAULTS_FILE_NAME
ENV_VAR_PREFIX = "JOBMON__"


class JobmonConfig:
    """Default config setup using YAML."""

    def __init__(self, filepath: str = "", dict_config: Optional[Dict] = None) -> None:
        """Jobmon config class.

        Args:
            filepath: where to read defaults from.
            dict_config: dictionary of values to override

        Config file priority:
            1. user specified file passed in
            2. environment variable JOBMON__CONFIG_FILE (backdoor for testing):q!
            3. config file from installer
            4. default config file in core
        """
        if filepath:
            self._filepath = filepath
        else:
            # Allow the user to specify a different config file using an environment variable
            self._filepath = os.getenv("JOBMON__CONFIG_FILE", "")
            # if the env not set, check if the installer plugin exists
            if self._filepath == "":
                # if the installer plugin exists, use the config file form the plugin
                self._filepath = CONFIG_FILE_FROM_INSTALLER_PLUGIN

        if self._filepath:
            with open(self._filepath, "r", encoding="utf-8") as f:
                self._config = yaml.safe_load(f)
        else:
            # when no config file in env and not installer plug-in,
            # use the default yaml in core
            self._filepath = DEFAULTS_FILE  # type: ignore
            with open(DEFAULTS_FILE, "r", encoding="utf-8") as f:
                self._config = yaml.safe_load(f)

        self._dict_config = dict_config

    def _merge_dicts(self, base: Dict, override: Dict) -> Dict:
        """Utility function to merge two dictionaries."""
        for key, value in override.items():
            if isinstance(value, dict):
                base[key] = self._merge_dicts(base.get(key, {}), value)
            else:
                base[key] = value
        return base

    def _get_env_var_name(self, section: str, key: str) -> str:
        return f"{ENV_VAR_PREFIX}{section.upper()}__{key.upper()}"

    def _get_environment_variable(self, section: str, key: str) -> Optional[str]:
        # must have format JOBMON__{SECTION}__{KEY} (note double underscore)
        env_var = self._get_env_var_name(section, key)
        return os.environ.get(env_var)

    def _interpolate_env_vars(self, value: Any) -> Any:
        if isinstance(value, str):
            return os.path.expandvars(value)
        return value

    def _get_yaml_variable(self, section: str, key: str) -> Optional[str]:
        return self._config.get(section, {}).get(key)

    def _get_dict_config_variable(self, section: str, key: str) -> Optional[str]:
        if self._dict_config:
            return self._dict_config.get(section, {}).get(key)
        return None

    def _wrapped_get(self, section: str, key: str) -> str:
        # First check in the dict_config
        val = self._get_dict_config_variable(section, key)
        if val is not None:
            return self._interpolate_env_vars(val)

        # Then check environment variable
        val = self._get_environment_variable(section, key)
        if val is not None:
            return val

        # Then check in the merged _config
        val = self._get_yaml_variable(section, key)
        if val is not None:
            return self._interpolate_env_vars(val)

        raise ConfigError(
            f'"{key}" key not found in "{section}" section of {self._filepath}. Fallback '
            f'option using environment var "{self._get_env_var_name(section, key)}" was not '
            "found."
        )

    def get(self, section: str, key: str) -> str:
        """Get the configuration value for the section and key. Raise if key not found.

        Args:
            section: the section of the yaml to search.
            key: the key within the section to retrieve

        Raises: ConfigError
        """
        try:
            val = self._wrapped_get(section, key)
            return self._interpolate_env_vars(val)
        except ConfigError as e:
            raise e

    def get_section(self, section: str) -> Dict[str, Any]:
        """Returns a dictionary of all key-value pairs in the given section.

        The order of precedence: dict_config > Environment Variable > YAML File.
        """
        # Start with the section from the YAML file, or an empty dict if the section
        # doesn't exist yet.
        section_dict = self._config.get(section, {}).copy()

        # Overlay values from dict_config, if any
        if self._dict_config:
            section_dict.update(self._dict_config.get(section, {}))

        # Check environment for variables related to this section and overlay them
        prefix = f"{ENV_VAR_PREFIX}{section.upper()}__"
        for env_key in os.environ.keys():
            if env_key.startswith(prefix):
                # Extract the original key name by removing the prefix
                key = env_key[len(prefix) :].lower()
                section_dict[key] = os.environ[env_key]

        return section_dict

    def get_boolean(self, section: str, key: str) -> bool:
        """Get the configuration value for the section and key as bool.

        Raise if key not found.
        """
        val = str(self.get(section, key)).lower().strip()
        if val in ("t", "true", "1", "yes"):
            return True
        elif val in ("f", "false", "0", "no"):
            return False
        else:
            raise ConfigError(
                f'Failed to convert value to bool. Please check "{key}" key in "{section}" '
                f'section or environment var "{self._get_env_var_name(section, key)}". '
                f'Current value: "{val}".'
            )

    def get_int(self, section: str, key: str) -> int:
        """Get the configuration value for the section and key as int.

        Raise if key not found.
        """
        val = self.get(section, key)
        try:
            return int(val)
        except ValueError as exc:
            raise ConfigError(
                f'Failed to convert value to int. Please check "{key}" key in "{section}" '
                f'section or environment var "{self._get_env_var_name(section, key)}". '
                f'Current value: "{val}".'
            ) from exc

    def get_float(self, section: str, key: str) -> float:
        """Get the configuration value for the section/key as float. Raise if key not found."""
        val = self.get(section, key)
        try:
            return float(val)
        except ValueError as exc:
            raise ConfigError(
                f'Failed to convert value to float. Please check "{key}" key in "{section}" '
                f'section or environment var "{self._get_env_var_name(section, key)}". '
                f'Current value: "{val}".'
            ) from exc

    def set(self, section: str, key: str, val: str) -> None:
        """Set the configuration value for the section/key."""
        if section not in self._config:
            self._config[section] = {}
        self._config[section][key] = val

    def write(self, filepath: Union[str, Path] = "") -> None:
        """Persist the current config to disk."""
        if not filepath:
            filepath = self._filepath
        if not filepath:
            filepath = DEFAULTS_FILE
        with open(filepath, "w") as f:
            yaml.safe_dump(self._config, f)


class ConfigCLI(CLI):
    """Client command line interface for update the config."""

    def __init__(self) -> None:
        """Initialization of client CLI."""
        super().__init__()
        self._subparsers = self.parser.add_subparsers(
            dest="sub_command", parser_class=argparse.ArgumentParser
        )
        self._add_update_config_subparser()

    @staticmethod
    def update_config(args: argparse.Namespace) -> None:
        """Update .jobmon.ini.

        Args:
            args: only --web_service_fqdn --web_service_port are expected.
        """
        config = JobmonConfig()
        config.set(
            "http",
            "service_url",
            f"http://{args.web_service_fqdn}:{args.web_service_port}",
        )
        config.write()

    def _add_update_config_subparser(self) -> None:
        update_config_parser = self._subparsers.add_parser("update")
        update_config_parser.set_defaults(func=self.update_config)
        update_config_parser.add_argument(
            "--web_service_fqdn", type=str, help="The fqdn of the web service."
        )
        update_config_parser.add_argument(
            "--web_service_port",
            type=str,
            help="The port for the web service..",
        )


def main(argstr: Optional[str] = None) -> None:
    """Create CLI."""
    cli = ConfigCLI()
    cli.main(argstr)



"""Task Instance object from the distributor's perspective."""

from __future__ import annotations

import logging
from typing import List, Set, Tuple, TYPE_CHECKING

from jobmon.core.constants import TaskInstanceStatus
from jobmon.core.requester import Requester

if TYPE_CHECKING:
    from jobmon.distributor.task_instance_batch import TaskInstanceBatch


logger = logging.getLogger(__name__)


class DistributorTaskInstance:
    """Object used for communicating with JSM from the distributor node."""

    def __init__(
        self,
        task_instance_id: int,
        workflow_run_id: int,
        status: str,
        requester: Requester,
    ) -> None:
        """Initialization of distributor task instance.

        Args:
            task_instance_id (int): a task_instance_id
            workflow_run_id (int): a workflow_run_id
            status(str): status of the distributor task instance
            requester (Requester, optional): a requester to communicate with
                the JSM. default is shared requester
        """
        self.task_instance_id = task_instance_id
        self.workflow_run_id = workflow_run_id
        self.status = status

        self.error_state = ""
        self.error_msg = ""

        self.requester = requester

    @property
    def submission_name(self) -> str:
        try:
            return self.batch.submission_name
        except AttributeError:
            return str(self.task_instance_id)

    @property
    def batch(self) -> TaskInstanceBatch:
        """Returns the batch the DistributorTaskInstance is in."""
        return self._batch

    @batch.setter
    def batch(self, val: TaskInstanceBatch) -> None:
        """Sets the batch of the DistributorTaskInstance."""
        self._batch = val

    @property
    def array_step_id(self) -> int:
        """Returns the array step of the TI."""
        return self._array_step_id

    @array_step_id.setter
    def array_step_id(self, val: int) -> None:
        self._array_step_id = val

    @property
    def distributor_id(self) -> str:
        """Returns the distributor id of the TI."""
        return self._distributor_id

    @distributor_id.setter
    def distributor_id(self, val: str) -> None:
        self._distributor_id = val

    def transition_to_launched(
        self, distributor_id: str, next_report_increment: float
    ) -> None:
        """Register the submission of a new task instance to a cluster.

        This method is never called by the happy path - only if array submission is not
        implemented on a particular cluster type.
        """
        self.distributor_id = distributor_id
        app_route = f"/task_instance/{self.task_instance_id}/log_distributor_id"
        self.requester.send_request(
            app_route=app_route,
            message={
                "distributor_id": str(distributor_id),
                "next_report_increment": next_report_increment,
            },
            request_type="post",
        )
        self.status = TaskInstanceStatus.LAUNCHED

    def transition_to_no_distributor_id(
        self,
        no_id_err_msg: str,
    ) -> None:
        """Register that submission failed with the central service.

        Args:
            no_id_err_msg: The error msg from the executor when failed to obtain distributor
                id.
        """
        app_route = f"/task_instance/{self.task_instance_id}/log_no_distributor_id"
        self.requester.send_request(
            app_route=app_route,
            message={"no_id_err_msg": no_id_err_msg},
            request_type="post",
        )

    def _transition_to_error(self, error_message: str, error_state: str) -> None:
        """Transitions the TaskInstance to the specified error state."""
        if self.distributor_id is None:
            raise ValueError("distributor_id cannot be None during log_error")
        distributor_id = self.distributor_id
        logger.debug(f"log_error for distributor_id {distributor_id}")
        if not error_state:
            raise ValueError("cannot log error if error_state isn't set")

        if error_state == TaskInstanceStatus.UNKNOWN_ERROR:
            app_route = f"/task_instance/{self.task_instance_id}/log_unknown_error"
        else:
            app_route = f"/task_instance/{self.task_instance_id}/log_known_error"

        self.requester.send_request(
            app_route=app_route,
            message={
                "error_state": error_state,
                "error_message": error_message,
                "distributor_id": distributor_id,
            },
            request_type="post",
        )
        self.error_state = error_state

    def transition_to_unknown_error(
        self, error_message: str, error_state: str
    ) -> Tuple[Set[DistributorTaskInstance], List]:
        """Register that an unknown error was discovered during reconciliation."""
        self._transition_to_error(error_message, error_state)
        return {self}, []

    def transition_to_resource_error(
        self, error_message: str, error_state: str
    ) -> Tuple[Set[DistributorTaskInstance], List]:
        """Register that a resource error was discovered during reconciliation."""
        self._transition_to_error(error_message, error_state)
        return {self}, []

    def transition_to_error(
        self, error_message: str, error_state: str
    ) -> Tuple[Set[DistributorTaskInstance], List]:
        """Register that a known error occurred during reconciliation."""
        self._transition_to_error(error_message, error_state)
        return {self}, []

    def __hash__(self) -> int:
        """Returns the id of the TaskInstance."""
        return self.task_instance_id

    def __eq__(self, other: object) -> bool:
        """Check if the hashes of two tasks are equivalent."""
        if not isinstance(other, DistributorTaskInstance):
            return False
        else:
            return hash(self) == hash(other)

    def __lt__(self, other: DistributorTaskInstance) -> bool:
        """Check if one hash is less than the has of another Task."""
        return hash(self) < hash(other)

    def __repr__(self) -> str:
        """Return a short representation string."""
        return (
            f"DistributorTaskInstance(task_instance_id={self.task_instance_id},"
            f"status={self.status})"
        )
